//*****************************************************************************
//
// File:          envelope.cc
// Purpose:       A simple linearly-interpolated envelope
//                implementation.  The routines are divided into two
//                classes, one that parses and stores the envelope
//                breakpoints and another that uses the parsed
//                envelope to perform the enveloping routine.  The
//                implementation is designed this way so that it is
//                most efficient for real-time applications.
//                Envelopes can be parsed and stored at startup time,
//                then, when the audio application is running, one of
//                the parsed envelopes can be used to envelope a
//                sound, no matter what it's length.  Once the
//                envelope is parsed, just connect it to a sound by
//                instantiating an envelope class, passing the
//                env-data class to the instance along with the number
//                of sample frames in the sound, and call the
//                envelope::env() routine.
// Author:        Michael Edwards
// Date:          4th April 1996
// Last revision: None
//
//*****************************************************************************


#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "envelope.h"

// The number of breakpoints to allocate memory for when reading the
// envelope.  If there's more breakpoints than this, another MAX
// breakpoints will be allocated.

static const int MAX = 32;


//*****************************************************************************

// Parse the envelope and store the values.  Memory for the break
// points is allocated dynamically so there's no limit on the number
// of breakpoints.  <env> is the envelope, which can be expressed in
// the form (x1 y1 x2 y2 x3 y3 ...) or (x1,y1,x2,y2,x3,y3 ...).
// Syntax errors will be reported if the envelope is not in this
// format.

env_data::env_data(char *env)
{
    char word[64];
    char *ptr = env;
    int i = 0;
    int word_count = 0;
    int OK = 0;
    float *output = new float[MAX];
    int next_alloc = MAX;

    // Duplicate <env>
    Envelope = new char[strlen(env)];
    strcpy(Envelope, env);

    // skip along to the left parenthesis
    while (*ptr++ != '(')
      ;
    do {

        // If the char is a number, a minus sign or a dot, store it.
        if (isdigit(*ptr) || *ptr == '-' || *ptr == '.') { 
            OK = 1;
            word[i++] = *ptr;
        }

        // Else it must be a space, ")", or a comma so get the value and
        // store it
        else if (OK && (*ptr == ',' || isspace(*ptr) || *ptr == ')')) {
            word[i] = '\0';
            output[word_count++] = float(atof(word)); // Put it in the array. 
            i = 0;
            OK = 0;  // Don't do the else part again until we've hit a digit.
        }

        // Else there's a syntax error, so print the envelope with a
        // "^" under the offending char.
        else {
            // The number of chars we've processed up to now.
            int x = ptr - env;

            cerr << "env_data::interpret_env():  Syntax error:\n" <<
              env << '\n';
            cerr.width(x + 1);
            cerr << '^' << '\n';
            exit(1);
        }

        // Dynamically allocate storage for the envelope values.
        if (word_count == next_alloc) {
            next_alloc += MAX;
            
            float *temp = new float[next_alloc];
            
            memcpy(temp, output, sizeof(float) * word_count);
            delete[] output;
            output = temp;
        }
    } while (*ptr++ != ')');

    // Allocate just enough storage for the values extracted and copy
    // them there.
    Values = new float[word_count];
    memcpy(Values, output, sizeof(float) * word_count);
    delete[] output;
    Data_count = word_count;
    check_env();
}


//*****************************************************************************

// Free memory allocated for <Values> and <Envelope>.

env_data::~env_data(void)
{
    delete[] Values;
    delete[] Envelope;
}

//*****************************************************************************

// Make sure that x values are all in ascending order, and that every
// x value has a corresponding y value.

void env_data::check_env(void)
{
    if (Data_count & 1) {
        cerr << "env_data::check_env():  Missing y value: \n"
          << Envelope << '\n'; 
        exit(1);
    }
    for (int i = 0; i < Data_count - 2; i += 2) {
        if (Values[i] > Values[i + 2]) {
            cerr << "env_data::check_env():  x values must be in "
              "ascending order:\n" << Envelope << '\n';
            exit(1);
        }
    }
}


//*****************************************************************************
//*****************************************************************************
//************************** envelope class methods ***************************
//*****************************************************************************
//*****************************************************************************


// Set the constants and initialize the private data members.  <Xinc>
// is the last x value in the envelope minus the first x value (which
// should be zero normally, but doesn't have to be) divided by the
// number of sample frames less one (so that we actually get to the
// last y value in the envelope).

envelope::envelope(const env_data &ed, const long num_frames) 
                   : Values(ed.getValues()),
                     Xinc((Values[ed.getData_count() - 2] - Values[0])
                          / (num_frames - 1))
{
    CurrentX = Values[0];
    X2 = CurrentX;

    // The first call to env() will result in next_segment() being
    // called so make sure that Index starts at the beginning
    // (i.e. the first call will result in Index being set to 0, which
    // is where we want to start).
    Index = -2;
}


//*****************************************************************************

// EOF envelope.cc
