//*****************************************************************************
//
// File:          envelope.h
// Purpose:       Header file for envelope classes.
// Author:        Michael Edwards
// Date:          8th April 1996
// Last revision: None
//
//*****************************************************************************


#include <iostream.h>


// Class to parse the envelope, check that it's OK (i.e. there's a Y
// value for every X value, and that the X values increase every time)
// and store the data so it can be accessed when called and passed to
// an envelope instance.

class env_data {

  public:

    // Constructor
    env_data(char *env);

    // Deconstructor
    ~env_data(void);

    // Accessor for <Values>.
    float *getValues(void) const { return Values; }

    // Accessor for <Data_count>.
    int getData_count(void) const { return Data_count; }

  private:
    
    // The break points.
    float *Values;

    // A copy of the envelope--used for diagnostic purposes.
    char *Envelope;

    // The number of values in the <Values> array, i.e. the number of
    // Xs and Ys in the envelope.
    int Data_count;

    // Function to check the envelope and make sure it's valid.
    void check_env(void);

    // There is no copy constructor. 
    env_data(const env_data &old_class);
};


//*****************************************************************************

// Class to perform linear interpolation of an envelope.  Here we pass
// the number of sample frames that the envelope will be stretched
// over.  The class uses the env_data instance to get the envelope data
// and perform the interpolation.

class envelope {

  public:
    
    // Constructor
    envelope(const env_data &ed, const long num_frames);

    // Accept default deconstructor.

    // The main envelope routine.  If we overstep the highest x value of
    // this segment, get the next segment, otherwise just increment
    // <CurrentY> and return it.  Inline'd here for efficiency's sake. 
    float env(void)
      {
          if (CurrentX >= X2)
              CurrentY = next_segment();
          else CurrentY += Yinc;
          CurrentX += Xinc;
          return CurrentY;
      }

  private:
    
    // The values taken from the env_data instance passed to the
    // constructor. 
    const float *Values;

    // The number of sample frames to stretch the envelope over.
    const long Num_frames;

    // The x-axis increment.
    const float Xinc;

    // The y-axis increment.
    float Yinc;

    // The second x value of the current segment.
    float X2;

    // The current y value.
    float CurrentY;

    // The current x value.
    float CurrentX;

    // The index into the <Values> array.  Incremented by two once we
    // have finished with each segment in turn.
    int Index;

    // Called when <CurrentX> oversteps the x limit of the current
    // segment.  We jump to the next set of breakpoints, linearly
    // interpolate the Y value to get the correct value for where we are
    // in the new segment, and set <Yinc> to the correct value for this
    // segment.  Inline'd for efficiency also.
    float next_segment(void)
      {
          Index += 2;
          
          const float X1 = Values[Index];
          const float Y1 = Values[Index + 1];
          
          X2 = Values[Index + 2];
          
          const float Y2 = Values[Index + 3];
          const float Xdiff = X2 - X1;
          
          Yinc = (Xdiff == 0.0) ? 0.0 : (Xinc * (Y2 - Y1)) / Xdiff;
          return (Y1 + (((CurrentX - X1) / (X2 - X1)) * (Y2 - Y1)));
      }

    // There is no copy constructor. 
    envelope(const envelope &old_class);
};


//*****************************************************************************

// EOF envelope.h
