//*****************************************************************************
//
// File:          machines.cc
// Purpose:       Audio and MIDI routines for supported machines
//                (currently SGI and NeXT).
// Author:        Michael Edwards
// Date:          December 13th 1996
// Last revision: None
//
//*****************************************************************************


#include "mix.h"


//*****************************************************************************
//*****************************************************************************
//******************** SGI audio and MIDI routines. ***************************
//*****************************************************************************
//*****************************************************************************

#ifdef SGI

// What percentage of the sound buffer to play before sending a new buffer to
// the SGI audio port. 

static const int FILLPERCENT = 90;
static ALport Audio_out_port;
static MDport Midi_in_port;


//*****************************************************************************

// Function prototypes:

static void underflow_error(void);
static ALport open_audio_port(void);

// Got to include this prototype because midi.h doesn't have the full ANSI
// style prototype. 

int mdInit(void);

//*****************************************************************************

// Continuously accept MIDI messages from the fader box and set the Volume
// associated with the appropriate channel.  <ignore> is there only because
// this process is forked using sproc() (on the SGI) and functions
// passed to sproc() require this argument.  Forked in mix::run(). 
 
void accept_volumes(void * /* ignore */)
{
    MDevent mdv;
    fd_set inports;
    fd_set *ptr;
    int nfds;
    int highfd;
    int channel;
    int volume;
    int nports;
    int i;

    // Do this silly operation to avoid a compiler warning caused by
    // the FD_SET() macro. 
    ptr = &inports;
    nports = mdInit();
    Midi_in_port = mdOpenInPort(0);
    if (Midi_in_port == NULL)
        error("accept_volumes", "Can't open MIDI input port.");
    mdSetStampMode(Midi_in_port, MD_NOSTAMP);
    FD_SET(mdGetFd(Midi_in_port), &inports);
    highfd = mdGetFd(Midi_in_port) + 1;
    for (;;) {
        
        // Block until select returns. 
        nfds = select(highfd, &inports, 0, 0, 0);
        for (i = 0; i < nports; i++) {
            if (FD_ISSET(mdGetFd(Midi_in_port), &inports)) {
                if (mdReceive(Midi_in_port, &mdv, 1) < 0)
                    fprintf(stderr, "Failure receiving message.");
                if (mdv.msg[0] == MD_SYSEX)
                    mdFree(mdv.sysexmsg);
            }
            else
                FD_SET(mdGetFd(Midi_in_port), &inports);
            channel = mdGetChannel(mdv.msg);
            volume = mdv.msg[2];

            // Wait for the semaphore. 
            GETSEMA();

            // Pass the values to the midi class.
            midi::setVolumes(channel, volume);

            // Release the semaphore to the other processes. 
            RELEASEMA();
        }
    }
}


//*****************************************************************************

// Check to see whether the audio port received samples fast enough. 
// One underflow error seems to be standard given the above
// audio_process() method.  It must come from opening the port when there
// aren't any samples to be sent quite yet. 
// Called by audio_process(). 
 

static void underflow_error(void)
{
    long pvbuf[4];

    pvbuf[0] = AL_ERROR_TYPE;
    pvbuf[2] = AL_ERROR_NUMBER;
    ALgetstatus(Audio_out_port, pvbuf, 4);
    
    if (pvbuf[1] == AL_ERROR_OUTPUT_UNDERFLOW) {
        Underflows = pvbuf[3];
    }
}

//*****************************************************************************

// Open an SGI audio port and return it.  Report appropriate errors if
// something goes wrong. Called by audio_process(). 

static ALport open_audio_port(void)
{
    ALport port;
    int err;
    ALconfig config = ALnewconfig();

    if (config == 0)
        error("open_audio_port",
              "Couldn't allocate a new audio configuration.");

    // Port is always stereo.  Make the queue size twice as big as we need. 
    if ((ALsetqueuesize(config, sound::getRead_frames() * 4)) == -1)
        error("open_audio_port",
              "Couldn't set queue size for the audio port.");
    port = ALopenport("mixport", "w", config);
    if (port == (ALport) 0) {
        err = oserror();
        if (err == AL_BAD_NO_PORTS)
            error("open_audio_port", "System is out of audio ports.");
        else if (err == AL_BAD_DEVICE_ACCESS)
            error("open_audio_port", "Couldn't access audio device.");
        else if (err == AL_BAD_OUT_OF_MEM)
            error("open_audio_port", "Out of memory: port open failed.");
    }
    ALfreeconfig(config);
    return port;
}


//*****************************************************************************


// Close the SGI MIDI and audio ports.  Called by mix::run().

void close_ports(void)
{
    ALcloseport(Audio_out_port);
    mdClosePort(Midi_in_port);
}



//*****************************************************************************

#endif    // SGI 


//*****************************************************************************
//*****************************************************************************
//*************************** NeXT MIDI routines. *****************************
//*****************************************************************************
//*****************************************************************************


#ifdef NEXT

extern "C" {
#include <mach/mach.h>
#include <stdio.h>
#include <mach/mach_error.h>
#include <servers/netname.h>
#include <libc.h>
#include <appkit/nextstd.h>
#include <mididriver/midi_driver.h>
#include <mididriver/midi_spec.h>

void playbuf(short *buf, int bufsize);
void init_sound(float sr, int chans);
}

static const int NO_RUNNING_STATUS = 0;
enum { FULL, FULLSYSEX, INCOMPLETE };


//*****************************************************************************

// Function prototypes. 
 

static void checkForError(char *msg, int errorReturn);
static port_t allocPort(void);
static void myDataReply(port_t replyPort, short unit,
			MIDIRawEvent * events, unsigned int count);
static port_t createPortSet(port_t dataPort);
static int midiParser(unsigned char aByte);
static unsigned char *setBufferSize(unsigned char *bytes, int size);
static void parse_message(unsigned char *bytes);
static void accept_volumes(void *ignore);



//*****************************************************************************

// NeXT equivalent of the SGI accept_volumes() function above.  Hacked
// from recordmidifile.c by Gregg Kellogg, Danna Massie and David 
// Jaffe.  
 

void accept_volumes(void *ignore)
{
    port_t driverPort;            // Port for driver on particular host. 
    port_t ownerPort;             // Port that represents ownership 
    port_t dataPort;              // Port for incoming data. 
    int unit = MIDI_PORT_A_UNIT;  // Serial port to read from 
    kern_return_t r;
    port_set_name_t ports;        // Port set to listen for messages from
                                  // driver 
    int synchUnit;
    MIDIReplyFunctions funcs = {0};

    // Set up MIDI driver 
    r = netname_look_up(name_server_port, "", "mididriver", &driverPort);
    checkForError("playmidifile: netname_look_up error", r);
    r = MIDIBecomeOwner(driverPort, ownerPort = allocPort());
    checkForError("MIDIBecomeOwner", r);
    r = MIDIClaimUnit(driverPort, ownerPort, unit);
    checkForError("MIDIClaimUnit", r);
    r = MIDISetClockMode(driverPort, ownerPort, synchUnit,
                         MIDI_CLOCK_MODE_INTERNAL);
    checkForError("MIDISetClockMode", r);
    r = MIDISetClockQuantum(driverPort, ownerPort, 1000);
    checkForError("MIDISetClockQuantum", r);
    ports = createPortSet(dataPort = allocPort());

    // Tell it to ignore system real time messages we're not interested in. 
    r = MIDISetSystemIgnores(driverPort, ownerPort, unit,
                             MIDI_IGNORE_REAL_TIME);
    checkForError("MIDISetSysIgnores", r);

    // Ask for data 
    r = MIDIRequestData(driverPort, ownerPort, unit, dataPort);
    checkForError("MIDIRequestData", r);

    // Request a message at time 1000ms (1 second). 
    r = MIDISetClockTime(driverPort, ownerPort, 0);
    checkForError("MIDISetClockTime", r);
    
    // We start clock now.  Alternatively, we could first queue up some
    // messages and then start time.  That would insure that the first few
    // notes come out correctly.  
    
    r = MIDIStartClock(driverPort, ownerPort);
    checkForError("MIDIStartTime", r);
    funcs.dataReply = myDataReply;
    while (1) {
        r = MIDIAwaitReply(ports, &funcs, MIDI_NO_TIMEOUT);
        checkForError("MIDIAwaitReply", r);
    }
}


//*****************************************************************************

// This gets invoked when data comes in. 

static void myDataReply(port_t replyPort, short unit, MIDIRawEvent * events,
                         unsigned int count)
{
    static int byteIndex;
    static int byteCount = 0;
    MIDIRawEvent *p;

    byteCount += count;
    for (p = events; count--; p++) {
        static unsigned char *bytes = NULL;

        bytes = setBufferSize(bytes, byteIndex + 1);
        bytes[byteIndex++] = p->byte;
        switch (midiParser(p->byte)) {
        case FULL:
            parse_message(bytes);
            byteIndex = 0;
            break;
        case FULLSYSEX:
            byteIndex = 0;
            break;
        default:
            break;
        }
    }
}


//*****************************************************************************


// Returns TRUE when a valid MIDI message is parsed.  We don't need to worry
// about System Real Time messages because they're filtered out by the
// driver. 
 

inline static int midiParser(unsigned char aByte)
{
    static unsigned char runningStatus = NO_RUNNING_STATUS;
    static int op = 0;
    static int nBytes = 0;
    static int msgSize = 0;
    int ret;

    if (aByte & MIDI_STATUSBIT) {
        if (aByte == MIDI_EOX && op == MIDI_SYSEXCL)
            return FULLSYSEX;
        op = aByte;
        if (MIDI_TYPE_SYSTEM(aByte))
            runningStatus = NO_RUNNING_STATUS;
        else
            runningStatus = op;
        msgSize = MIDI_EVENTSIZE(runningStatus);
        nBytes = 0;
    }
    ret = ((op == MIDI_SYSEXCL) ? INCOMPLETE :
           (++nBytes == msgSize) ? FULL : INCOMPLETE);
    if (ret == FULL)
        nBytes = 1;
    return ret;
}


//*****************************************************************************

inline static unsigned char *setBufferSize(unsigned char *bytes, int size)
{
    static int bufsize = 0;

    if (!bytes)
        NX_MALLOC(bytes, unsigned char, bufsize = size);
    else if (size < bufsize) {
        bufsize = size * 2;
        NX_REALLOC(bytes, unsigned char, bufsize);
    }
    return bytes;
}


//*****************************************************************************

// Get the channel and volume values out of a full MIDI message. The type of
// MIDI message we will be parsing will consist of two or three bytes.  The
// volume value is always the last byte in the message.  If there are three
// bytes, then the first contains channel info (it's the status byte and
// channel is in the least significant four bits of the byte, hence we
// bitwise-AND it with 0x0f to get the channel). 
 
inline static void parse_message(unsigned char *bytes)
{
    static int ch = -1;
    int vol = 127;

    if (*bytes & MIDI_STATUSBIT) {
        ch = *bytes & 0x0f;
        bytes++;
    }
    vol = *++bytes;
    GETSEMA();

    // Pass the values to the midi class.
    midi::setVolumes(ch, vol);
    RELEASEMA();
}


//*****************************************************************************

// Checks for error.  If error, prints message and quits. 

static void checkForError(char *msg, int errorReturn)
{
    if (errorReturn != KERN_SUCCESS) {
        switch (errorReturn) {
            case MIDI_ERROR_BUSY:
             printf("%s: %s", msg, "MIDI driver busy.\n");
            break;
        case MIDI_ERROR_NOT_OWNER:
            printf("%s: %s", msg, "You must be owner of the MIDI driver.\n");
            break;
        case MIDI_ERROR_QUEUE_FULL:
            printf("%s: %s", msg, "MIDI driver queue full.\n");
            break;
        case MIDI_ERROR_BAD_MODE:
            printf("%s: %s", msg, "Bad MIDI driver clock mode.\n");
            break;
        case MIDI_ERROR_UNIT_UNAVAILABLE:
            printf("%s: %s", msg, "MIDI driver unit unavailable.\n");
            break;
        case MIDI_ERROR_ILLEGAL_OPERATION:
            printf("%s: %s", msg, "MIDI driver illegal operation.\n");
            break;
        default:
            mach_error(msg, errorReturn);
        }
    }
}


//*****************************************************************************

// Allocates a port and returns it. 

static port_t allocPort(void)
{
    port_t aPort;
    int r = port_allocate(task_self(), &aPort);

    checkForError("allocPort", r);
    return aPort;
}


//*****************************************************************************

// Creates the port set and adds the three ports.  

static port_t createPortSet(port_t dataPort)
{
    port_set_name_t aPortSet;
    int r = port_set_allocate(task_self(), &aPortSet);

    checkForError("createPortSet", r);
    r = port_set_add(task_self(), aPortSet, dataPort);
    checkForError("createPortSet", r);
    return aPortSet;
}


//*****************************************************************************

#endif  // NEXT 


//*****************************************************************************
//*****************************************************************************
//************************* Main audio loop for NeXT and SGI ******************
//*****************************************************************************
//*****************************************************************************


//*****************************************************************************

// Open a port and write the samples returned by soundlist::mix() out to it.
// <the_soundlist> is a void pointer because this process is forked via
// sproc() (on the SGI) and the only thing this function can pass is a
// pointer of this type. Forked in mix::run(). 
 

typedef short *(soundlist::*mix_fun_ptr)(void);

void audio_process(void *the_soundlist)
{
    soundlist *sndlist;
    // Pointer to the mix sample buffer. 
    short *mixptr;
    // The mix buffer is always stereo. 
    const long num_samples = sound::getRead_frames() * 2;

#ifdef SGI
    // How much to let the audio port drain before sending new samples to
    // it: 
    long lowpoint = (sound::getRead_frames() / 100) * FILLPERCENT;

    // Try to get the machine to run with a higher priority.  Only
    // succeeds if you're running as root.
    if (schedctl(NDPRI, 0, NDPNORMMAX) < 0) {
      // fprintf(stderr, "Run as root for higher process priority.\n\r");
      // fflush(stderr);
    }
    if (prctl(PR_RESIDENT) == -1) {
      // fprintf(stderr, "Can't prevent memory swapout.\n\r");
      // fflush(stderr);
    }
    Audio_out_port = open_audio_port();
#endif //SGI

#ifdef NEXT
    init_sound(sound::getSampling_rate(), 2);
#endif // NEXT

    mix_fun_ptr mixfun = (Fast ?
                          &soundlist ::simple_mix:&soundlist::mix_and_limit);

    // (re)cast the sound list: 
    sndlist = (soundlist *)the_soundlist;

    // Fill all the sounds' buffers that are in the list: 
    sndlist->fillbuffers();

    // Mix all the buffers: 
    mixptr = (sndlist->*mixfun)();

    //********************* Main audio loop ********************************

    // This process is killed in mix::run() so the loop goes forever. 
    while (1) {

#ifdef LITTLE_ENDIAN
        sound::swap_buf(mixptr);
#endif

        // Play the mix: 

#ifdef SGI
        ALwritesamps(Audio_out_port, mixptr, num_samples);
#endif //SGI

#ifdef NEXT
        playbuf(mixptr, num_samples);
#endif // NEXT

        sndlist->fillbuffers();
        
        // Wait for the semaphore so we don't set any sound object's fields
        // while other processes are working on them: 
        GETSEMA();
        sndlist->tidy();

        // Release the semaphore to the other processes. 
        RELEASEMA();
        mixptr = (sndlist->*mixfun)();

        // Only print underflow errors if the user has told us to (on the
        // command line). 
        if (!Silent && OK_to_print) {

#ifdef SGI
            underflow_error();
#endif // SGI
	    fprintf(stderr, "%140s \r", "");
	    if (TestMidi)
	      fprintf(stderr,"      %03d %03d %03d %03d %03d %03d %03d "
		      "%03d %03d %03d %03d %03d %03d %03d %03d %03d    ",
		     midi::Volumes[0], midi::Volumes[1],
		     midi::Volumes[2],
		     midi::Volumes[3], midi::Volumes[4], midi::Volumes[5],
		     midi::Volumes[6], midi::Volumes[7], midi::Volumes[8],
		     midi::Volumes[9], midi::Volumes[10], midi::Volumes[11],
		     midi::Volumes[12], midi::Volumes[13], midi::Volumes[14],
		     midi::Volumes[15]);
            fprintf(stderr, "%-12ld%-12ld%-2d", Underflows, Limits,
		    sndlist->getSounds_playing());
            sndlist->printkeys();
	    putc('\r', stderr);
	}
            
#ifdef SGI       
        // Wait until FILLPERCENT percent of the samples have been played
        // before filling the buffers and mixing again: 
        while (ALgetfilled(Audio_out_port) > lowpoint)
            sginap(1);
#endif // SGI
      
    }
}


//*****************************************************************************

// EOF machines.cc 
 
