//*****************************************************************************
//
// File:          mix.h
// Purpose:       Header file included in every .cc file in the mixing
//                programme.
// Author:        Michael Edwards
// Date:          30th September 1995
// Last revision: None
//
//*****************************************************************************

#include <iostream.h>
#include <math.h>
#include <stdlib.h>

#ifdef NEXT
//extern void swab(char *from, char *to, int nbytes);
#include <curses.h>
extern "C" {
#include <libc.h>
#include <mach/cthreads.h>
}
#endif  // NEXT 

#ifdef SGI
#include <errno.h>
#include <unistd.h>
#include <bstring.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/schedctl.h>
#include <sys/select.h>
#include <ulocks.h>
extern "C" {
#include <curses.h>
}
#include <dmedia/audio.h>
#include <dmedia/midi.h>
#endif  // SGI 


// Macros to handle semaphores in forked processes. 

#ifndef   DEBUG
#ifdef    SGI
#define   GETSEMA() uspsema(Semaphore)
#define   RELEASEMA() usvsema(Semaphore)
#endif    // SGI 

#ifdef    NEXT
#define   GETSEMA() mutex_lock(Semaphore)
#define   RELEASEMA() mutex_unlock(Semaphore)
#endif    // NEXT 
#else     // We're in debug mode, so no semaphore control
#define GETSEMA()
#define RELEASEMA()
#endif    // DEBUG

#define   MAXPATH    128  // Max. number of chars in a path name. 
#define   MAXLINE    256  // Max. length of a line in the mappings file. 
#define   MAXSOUNDS   58  // Max. number of sounds that can be mapped.


typedef enum {
    NO, YES
} BOOL;


//*****************************************************************************

// externs for globals. 

extern BOOL Silent;
extern BOOL Fast;
extern long Limits;
extern long Underflows;
extern BOOL OK_to_print;
extern BOOL Info;
extern BOOL TestMidi;
extern double Expt;

#ifdef SGI
extern usema_t *Semaphore;
#endif // SGI

#ifdef   NEXT
extern mutex_t Semaphore;
#endif // NEXT



//*****************************************************************************
//
// sndfile is a class that automatically reads sound file headers and
// stores the information necessary for playing/processing the sounds.
// When declaring an instance of this class, you pass the full path of the
// sound file and the constructor takes care of getting all the data.  At
// the moment, the only supported sound file formats are AIFF and
// SND--these are read transparently to the user, i.e. you don't have to
// specify which format you are using.  If any of the data is not valid
// for use, an error will be signalled (for instance if the file is
// neither AIFF nor SND, or if the samples are not 16-bit linear.
// Note that after the header has been read, the file pointer is neither
// reset to the beginning, nor to point to the start of the samples.  This
// must be done before accessing the samples.
//
//*****************************************************************************

class sndfile {

  public:

    // Constructor. 
    sndfile(const char *path);

    // Destructor. 
    virtual ~sndfile(void);

    // Accessor for <Channels>.
    int getChannels(void) const { return Channels; }

    // Accessor for <Fp>.
    FILE *getFp(void) const { return Fp; }

  protected:

    // Where the samples actually start in the file. 
    int Data_location;

    // Number of channels in the sound file. 
    int Channels;

    // Number of sample **frames** (not samples) in the sound file. 
    int Num_frames;

    // The sampling rate of the sound.  Only 22.05 and 44.1 kHz are legal. 
    int Srate;

    // The file pointer used to access the samples. 
    FILE *Fp;

  private:

    // Reader for AIFF files (Mac, SGI). 
    void read_aiff(const char *path);

    // Reader for SND files (NeXT, Sun). 
    void read_snd(const char *path);

    // Conversion function for the awkward sampling rate field of AIFF files.
    double ieee_80_to_double(char *p);

    // Function used in ieee_80_to_double(). 
    inline double myUlongToDouble(unsigned long ul);

    // Return a four-byte int from a char pointer. 
    inline int get_int(char *start);

    // Return a two-byte int from a char pointer. 
    inline short get_short(char *start);

    // Place to store HDRSIZE number of chars for reading the header. 
    char *Hdrbuf;

    // There is no copy constructor. 
    sndfile(const sndfile &old_class);
};


//*****************************************************************************
//
// Simple class to parse and store the MIDI channel and scaler
// information for the MIDI fader box values.  In the mappings text
// file the MIDI channel can be specified as , e.g., "midich = 12".
// This means that the amplitude (or whatever) value will be taken
// from the fader that sends messages on channel 12, and the result
// will be scaled from 0-127 down to 0.0-1.0 (the default).  Or the
// MIDI channel could be specified as "midich = 12:5.0:7.9" which
// would be interpreted to mean receive messages from channel 12, but
// instead of scaling down to 0.0-1.0, scale from 5.0-7.9.  The value
// returned by midi::get_dvalue will always be in this range.  By
// calling this function a double value will be returned.  By calling
// midi::get_ivalue, an integer value (rounded to the nearest int)
// will be returned instead.
//
//*****************************************************************************

class midi {

  public:

    // Constructor.
    midi(void);

    // Accept default destructor.

    // Accessor for <Channel>.
    int getChannel(void) const { return Channel; }

    // Accessor for <High>.
    double getHigh(void) const { return High; }

    // Accessor for <Low>.
    double getLow(void) const { return Low; }

    // Accessor for <Last_amp>.
    double getLast_amp(void) const { return Last_amp; }

    // Parse the colon separated fields into channel, high and low values.
    void parse(char *data);

    // Return the value from the <Volumes> array, scaled appropriately so
    // that it falls between <Low> and <High>, and raised to the power of
    // <Expt> as supplied on the command line.  Just return 1.0 if the
    // (default) channel is 0.  This function is called every loop
    // through the audio process so define it here so it's called
    // inline by sound::scale().
    // Side-effect is that <Last_amp> gets set here to whatever this
    // function will return.
    double get_dvalue(void)
      {
          return (Last_amp = 
                  (Channel == 0 ? 1.0 :
                   Low + ((High - Low)
                          * pow(double(Volumes[Channel - 1]) / 127.0, Expt))));
      }

    // Return the integer value from the <Volumes> array.  Rounds a
    // double value to the nearest int.  Inline this one also.
    int get_ivalue(void) { return int(rint(get_dvalue())); }

    // Set the volume value for channel <channel> in <Volumes>.
    static void setVolumes(int channel, int value)
      {
          if (channel > 16 || value > 127 || value < 0)
            cerr << "setVolumes:  Invalid values received from MIDI:  "
              "Channel: "  << channel << " Value: " << value << '\n';
          else {
            Volumes[channel] = value;
          }
      }

    // Initialize all the channels' volumes to 0 (minimum) so that all
    // sounds are played at no amplitude until one of the faders is moved. 
    static void initVolumes(void)
      {
          for (int i = 0; i < 16; i++)
            Volumes[i] = 127;
      }
    
  private:

    // The last amplitude that was sent by the MIDI fader box.  Used to
    // interpolate between the values used to scale each successive buffer of
    // samples sent to the DAC. 
    double Last_amp;

    // The line to be parsed, containing colon separated fields for
    // channel, highest value and lowest value.
    char *Data;

    // Array to hold volume values read from the MIDI fader box.  16 channels
    // total.  Set in accept_volumes(). 
public:
    static int Volumes[16];

private:
    // The midi channel to receive messages on.
    int Channel;

    // The highest value to return--used to scale the 0-127 volume
    // message. 
    double High;

    // The lowest value to return.
    double Low;

    // Get the next colon-separated field.  Called by parse.
    BOOL split(char *holder, int limit);

    // There is no copy constructor. 
    midi(const midi &old_class);
};


//*****************************************************************************

// Class to hold and process the parameters specified in the key mappings
// text file. 

class parameters {

  public:

    // Constructor.  <line> is the parameter info, <pathname> is the
    // directory of where the sounds are to be found, and <line_num>
    // is the line number in the mappings file (used for diagnostic
    // purposes). 
    parameters(char *line, char *pathname, int line_num);

    // Destructor. 
    virtual ~parameters(void);

    // Accessor for <Key>.
    char getKey(void) const { return Key; }

    // Accessor for <Path>.
    char *getPath(void) const { return Path; }

    // Accessor for <Restart>.
    BOOL getRestart(void) const { return Restart; }

    // Accessor for <Src>.
    double getSrc(void) const { return Src; }

    // Accessor for <Reverse>.
    BOOL getReverse(void) const { return Reverse; }

  protected:

    // The key the sound is mapped to. 
    char Key;
    
    // The full path of the sound. 
    char *Path;
    
    // The midi class that processes the volume value used to scale
    // the sound's amplitude. 
    midi Midich;
    
    // How many seconds into the sound to start playing. 
    double Skip;
    
    // A global scaler for the sound's amplitude.  Multiplied by the MIDI
    // channel velocity in the scaling function. 
    double Amp;

    // Whether to play the sound backwards or not. 
    BOOL Reverse;
    
    // Whether hitting the key when the sound is playing restarts the sound
    // or stops it. 
    BOOL Restart;

    // The sampling rate conversion. 
    double Src;

  private:

    // String to store the values of the parameters in.  Allocated in the
    // constructor and freed at the end of the constructor. 
    char *Value;
    
    // The line pointer, incremented as we process each parameter.  We need
    // this extra pointer just in case there's an error, so that we can print
    // the whole line and not just a portion of it. 
    char *Lineptr;

    // Function to test whether the parameter given is legal or not. 
    int legal(char *param);

    // Function to get the next parameter. 
    int getparam(void);

    // There is no copy constructor. 
    parameters(const parameters &old_class);
};


//*****************************************************************************


// Class to hold the necessary data and functions for playing and
// manipulating a sound file. 

class sound: public parameters, public sndfile {

  public:
    
    // Constructor. 
    sound(char *params, char *pathname, int line_num);

    // Destructor. 
    ~sound(void);

    // Accessor for <Playing>.
    BOOL getPlaying(void) const { return Playing; }

    // Mutater for <Playing>.
    void setPlaying(BOOL val) { Playing = val; }

    // Accessor for <Samples>.
    short *getSamples(void) const { return Samples; }

    // Accessor for <Frames_left>.
    long getFrames_left(void) const { return Frames_left; }

    // Accessor for <Start_frame>.
    long getStart_frame(void) const { return Start_frame; }

    // Accessor for <Read_frames>.
    static long getRead_frames(void) { return Read_frames; }

    // Accessor for <Sampling_rate>.
    static long getSampling_rate(void) { return Sampling_rate; }

    // Allocate memory for <Tempbuf>.
    static void initTempbuf(long siz);

    // Free memory allocated for <Tempbuf>.
    static void freeTempbuf(void);

    // Tell the sound it's not playing and fseek() to <Start_byte>. 
    void reset(void);
    
    // Print sound statistics in long or short form depending on the
    // value of <Info>. 
    void printStats();

    // Reverse a buffer of samples to play a sound backwards. 
    void reverse_buffer(void);
    
    // Convert a buffer from mono to stereo by simple duplication of every
    // sample. 
    void mono_to_stereo(void);
    
    // Scale the samples by the product of <Amp> and the value sent by the
    // MIDI fader box. 
    void scale(void);

    BOOL JustStarted;

    BOOL JustStopped;

  private:

    // The number of **frames** (not samples) to read into each
    // sound's sample buffer. Set in sound::sound(). 
    static long Read_frames;

    // The sampling rate we'll be running the programme at.
    static long Sampling_rate;

    // The place where we temporarily store the samples that are to be
    // processed.  Sounds are processed sequentially so we only need
    // allocate one temporary buffer for the whole class, rather than
    // one per instance.
    static short *Tempbuf;

    static float* MonoRamp;

    static float* StereoRamp;

    // Number of frames left to read.  Initialized to <num_frames>.
    long Frames_left;

    // The sample to start at (default 0). 
    long Start_frame;

    // Byte number to fseek() to. 
    long Start_byte;

    // The buffer to be filled with samples. 
    short *Samples;

    // The sound's duration in seconds. 
    double Duration;

    // Whether the sound file is being played or not. 
    BOOL Playing;

#ifdef   SGI
    // Functions to set hardware parameters on the Indy. 
    long GetOutputRate(void);
    long GetInputRate(void);
    void setoutputrate(long srate);
#endif // SGI

    // There is no copy constructor. 
    sound(const sound &old_class);

    void MakeRampDownBuf(float* buf, long size);

    // Need to put more public function definitions down here so
    // that they know the existence of <Read_frames> (gcc
    // complains otherwise).

  public:

    // Swap bytes on LITTLE_ENDIAN machines.  Swaps a full buffer of
    // <Read_frames> * 2 (for stereo) samples as a default.
    static void swap_buf(short *samples, int nbytes =
                         Read_frames * 2 * sizeof(short))
      {
          swab((char *)samples, (char *)Tempbuf, nbytes);
          bcopy((void *)Tempbuf, (void *)samples, nbytes);
      }

    // Read a buffer's worth of samples from the sound file into <Samples>. 
    long readsamples_aux(long num, BOOL samples);
    long readsamples(long num = Read_frames);
};


//*****************************************************************************


// Class to keep track of what sounds are playing and for mixing them
// together. 

class soundlist {

  public:

    // Constructor. 
    soundlist(void);

    // Destructor. 
    ~soundlist(void);

    // Accessor for <Sounds_playing>.
    int getSounds_playing(void) const { return Sounds_playing; }

    // Add a sound to the list. 
    void add(sound *new_sound);

    // Remove all the sounds that are not currently playing. 
    void tidy(void);

    // Stop all the currently playing sounds. 
    void stopall(void);

    // Print the keyboard keys associated with the currently playing sounds. 
    void printkeys(void);
    
    // Mix the currently playing sounds and make sure they don't wrap around
    // the 16-bit limit. 
    short *mix_and_limit(void);
    
    // Mix the currently playing sounds but don't bother checking for
    // wraparound. 
    short *simple_mix(void);

    // Fill all the <Samples> buffers of all the currently playing sounds. 
    void fillbuffers(void);

  private:

    class soundlist_element {
      
      public:
        
        // The sound data. 
        sound *Snd;
        
      private:
        
        // Next in the list. 
        soundlist_element *Next;
        
        // Do this so we can access <Next>. 
    friend class soundlist;
    };

    // The number of sounds in the list that are currently playing. 
    int Sounds_playing;
    
    // A buffer filled with silence, to be returned by mix() when no sounds
    // are in the list. 
    short *Silence;

    // The buffer returned by mix(). 
    short *Mixbuf;

    // The buffer we mix into before limiting the samples (with
    // mix_and_limit()). 
    long *Tempmixbuf;

    // The start of the list. 
    soundlist_element *Head;

    // There is no copy constructor. 
    soundlist(const soundlist &old_class);
};


//*****************************************************************************


// Class that gets the whole programme running and keeps control of
// the data. 
// N.B.  The actual mixing algorithms are in the soundlist class, not
// here.  

class mix {

  public:

    // Constructor.  <file> is the mappings file.
    mix(char *file);

    // Destructor. 
    ~mix(void);

    // The main key catching and processing routine. 
    void run(char *timestampfile);

  private:

    // Array to hold the key mappings of all the sound files. 
    sound *Mappings[MAXSOUNDS];

    // Write the timings of all the key strokes to a file. 
    void timestamp(FILE *fp, sound *sr, BOOL start);

    // Get user input from the keyboard (i.e. type in SRC values etc.). 
    void get_input(int key);
};


//*****************************************************************************


// Function prototypes: 
 

// Simple error function that causes exit. 
void error(const char *function_name, const char *format, ...);

// Trim whitespace from either end of a string. 
void stringtrim(char *str);

// Convert semitone transposition to sampling-rate conversion factor. 
double st_to_src(char *val);

// The main audio loop involving writing mixed samples to the DAC. 
void audio_process(void *the_soundlist);

// The MIDI process that gets fader volumes and stores them. 
void accept_volumes(void *ignore);

// Routine to get the filename minus extension from any full path. 
void filename(char *path, char *name);

// Close audio and MIDI ports on the SGI, free allocated sound
// resources on the NeXT.
extern "C" {
void close_ports(void);
}


//*****************************************************************************

// EOF mix.h 
 
