//*****************************************************************************
//
// File:     parameters.cc
// Purpose:  Take a line from the mappings text file and parse all the
//           arguments in the line, returning a parameters class
//           with all the relevant fields filled in by the values in the line.
//           Values are specified using an "=" sign and are separated by
//           commas.  Arguments can appear in any order.
//           Legal parameters are specified in the Parameters global array.
//           The parameters class is initialized with reasonable defaults so
//           all a line has to contain is the key and the sound.
//           E.g.  amp = .6, key = q, snd = rain.snd, midich = 5
// Author:   Michael Edwards
// Date:     5th November 1995
// Last revision: None
//
//*****************************************************************************


#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <iostream.h>
#include "mix.h"


//*****************************************************************************

// Global return values for the parsing mechanism. 
 

static const int ILLEGAL = -1;
static const int ENDLINE = -2;


// Array to hold the legal parameters that a line can contain.  To add more,
// add the new parameter after the last in the array, add a new case test to
// the switch in parameters::parameters() and add the new field to the
// parameters class, giving a reasonable default in parameters::parameters().
 

static const char *Parameters[] = {"key", "snd", "midich", "skip", "amp",
				   "reverse", "src", "restart"};


//*****************************************************************************

// Constructor. Parse all the arguments in the line.  <line_num> is only
// needed for error reporting.  Returns the parameter class with all the
// appropriate fields filled in by the values in <line>.  Called in
// mix::mix(). 
 
parameters::parameters(char *line, char *pathname, int line_num)
{
    BOOL got_key = NO;
    BOOL got_path = NO;
    BOOL keep_going = YES;

    // Set reasonable defaults.  No need to set the Midich defaults as
    // they are set automatically in the constructor for the midi
    // class. 
    Key = 0;
    Path = "";
    Skip = 0.0;
    Amp = 1.0;
    Reverse = NO;
    Restart = NO;
    Src = 1.0;
    Lineptr = line;
    Value = new char[MAXLINE];

    while (keep_going) {
	switch (getparam()) {
	case 0:
	    Key = *Value;
	    got_key = YES;
	    break;
	case 1:
	    Path = new char[strlen(Value) + strlen(pathname) + 1];
	    strcpy(Path, pathname);
	    strcat(Path, Value);
	    got_path = YES;
	    break;
	case 2:
	    Midich.parse(Value);
	    break;
	case 3:
	    Skip = atof(Value);
	    break;
	case 4:
	    Amp = atof(Value);
	    break;
	case 5:
	    
	    // We don't care what value is assigned to <Reverse>:  if it is
	    // mentioned in the line, <Reverse> will be set, so do
	    // nothing with  <Value>. 
	    Reverse = YES;
	    break;
	case 6:
	    Src = st_to_src(Value);
	    break;
	case 7:
	    
	    // Similar to <Reverse>.
	    Restart = YES;
	    break;
	case ILLEGAL:
	    
	    // Report the syntax error giving the line number and placing a
	    // "^" under the offending parameter. 
	    error("parameters::parameters",
		  "Syntax error at line %d:\n\r%s\n\r%*s",
		  line_num, line, (1 + strlen(line)) - (strlen(Lineptr)), "^");
	case ENDLINE:
	    keep_going = NO;
	}
    }
    if (!got_key || !got_path)
	error("parameters::parameters",
	      "Syntax error at line %d:\n\rYou must specify at least a key"
	      " and a sound:\n\r%s", line_num, line);
    delete[] Value;
}


//*****************************************************************************

// Destructor. 

parameters::~parameters(void)
{
    delete[] Path;
}


//*****************************************************************************

// Get the value of the parameter.  First check that the parameter is legal,
// then place the value of the parameter in the array <val> so that it can be
// appropriately converted by the caller.  Returns the position of the
// parameter within the Parameters array so that we can switch on this value.
// Called by parameters::parameters(). 

int parameters::getparam(void)
{
    char buf[MAXPATH];		// Storage for the processed characters. 
    int i = 0;			// Character counter. 
    int result;			// The position within the Parameters array. 

    // Get the parameter name: 
    while (*Lineptr != '=' && *Lineptr != ',' && *Lineptr != '\0')
	buf[i++] = *Lineptr++;
    buf[i] = '\0';
    stringtrim(buf);
    if (*buf == '\0')
	return ENDLINE;	// We got to the end of the line. 
    
    // Get the array reference of the parameter or return if the given
    // parameter is not in the Parameters array: 
    if ((result = legal(buf)) == ILLEGAL) {

	// Set the pointer back to the beginning of the offending parameter
	// so that we can print the "^" in the error message in the right
	// place. 
	Lineptr -= i;
	return ILLEGAL;
    }
    
    // In the case of <Reverse> being specified, we don't need to give it a
    // value, so there will be no equals sign and we don't need to proceed,
    // so just return <result> so we can switch and act accordingly. 
    if (*Lineptr == ',') {
	// Skip the comma so we can get the next parameter. 
	++Lineptr;
	return result;
    }
    ++Lineptr;			// Skip equals sign. 

    // Reset i so we can now fill <buf> with the parameter value: 
    i = 0;

    // Get the value until we hit a comma or end of string: 
    while (*Lineptr != ',' && *Lineptr != '\0')
	buf[i++] = *Lineptr++;
    buf[i] = '\0';
    stringtrim(buf);

    // Store the value 
    strcpy(Value, buf);
    
    // Skip the comma and whitespace (stringtrim takes care of this but we
    // need to skip whitespace so that syntax errors can be reported in a
    // useful way). 
     
    while (*Lineptr == ',' || isspace(*Lineptr)) {
	++Lineptr;
    }
    return result;
}


//*****************************************************************************

// Return the position of <param> within the Parameters array or "ILLEGAL" if
// it's not there.  Called by parameters::getparam(). 

int parameters::legal(char *param)
{
    // Get the number of elements in Parameters[]: 
    static int num = sizeof Parameters / sizeof Parameters[0];
    int i;

    for (i = 0; i < num; i++)
	if (!strcmp(param, Parameters[i]))
	    return i;
    return ILLEGAL;
}


//*****************************************************************************



//*****************************************************************************
//*****************************************************************************
//
// End of parameters class methods.
//
//*****************************************************************************
//*****************************************************************************



//*****************************************************************************

// Return a sampling-rate conversion factor from a string. <Value> is the
// factor to be converted.  If it represents a semitone transposition, the
// first two characters must be "st", otherwise it is assumed to be a simple
// sampling-rate conversion factor. 

double st_to_src(char *val)
{
    if (!strncmp(val, "st", 2)) {
	 val += 2;
	return pow(2.0, atof(val) / 12.0);
    }
    else
	return atof(val);
}



//*****************************************************************************

// Simple error handling function.  Forces exit of programme. Works exactly
// like any of the printf family, except that it prints an introductory
// message supplying the function (<function_name>) that signalled the error
// as well as the user supplied message with all the usual formatting
// characteristics. 
 
void error(const char *function_name, const char *format,...)
{
    va_list args;

    va_start(args, format);
    fprintf(stderr, "Error signalled by function %s():\n\r", function_name);
    vfprintf(stderr, format, args);
    fprintf(stderr, "\n\r");
    va_end(args);
    exit(1);
}

//*****************************************************************************

// Trim leading and trailing white space from a string. N.B.  This is a
// **destructive** routine.  <str> is the string we want to trim. 

void stringtrim(char *str)
{
    char *ptr = str;

    // Return if empty string.
    if (*ptr == '\0')
	return;

    // Skip whitespace at beginning.
    while (*ptr != '\0' && isspace(*ptr++))
      ;

    // Copy the pointer back into <str>.
    strcpy(str, --ptr);

    // Go to the end of the string. 
    ptr += strlen(ptr);

    // Go backwards until we hit a non-whitespace character.
    while (isspace(*--ptr))
      ;

    // Insert a string terminator.
    *++ptr = '\0';
}

//*****************************************************************************

// Place the name of a file in <name>, stripping off the extension so that
// e.g. "/usr/people/michael/my.sound.aiff" becomes "my.sound".
// N.B.  Doesn't check that name is big enough to strcpy into. 
 
void filename(char *path, char *name)
{
    int i = 0;

    // Go to the end of the path. 
    path += strlen(path) - 1;

    // Go back until we hit the last dot (start of extension) 
    while (*path != '.')
	--path;

    // Go back to the last slash and count how many chars. 
    while (*path != '/') {
	--path;
	++i;
    }

    // Went back one char too many. 
    ++path;

    // i is now one more than we need but make room for the '\0'. 
    strncpy(name, path, i - 1);
    name[i - 1] = '\0';
}


//*****************************************************************************

// EOF parameters.cc 
 
