//*****************************************************************************
//
// File:          src.cc
// Purpose:       Sample rate convert (transpose) a sound file filling a sound
//                object's sample buffer with converted samples.
// Author:        Michael Edwards
// Date:          9th November 1995
// Last revision: 5th December 1995
//
//*****************************************************************************

#include <iostream.h>
#include <math.h>
#include <string.h>
#include <bstring.h>
#include "mix/mix.h"

//*****************************************************************************
//
// The code below is a hacked version of Perry Cook's SRC programme.
// The conversion algorithm is 100% Perry.  My only changes were to
// implement the routines within a C++ class framework and adjust the
// procedures so that they work with buffers of samples instead of doing
// a whole sound file in one go.
// Many thanks to Perry for allowing me to use this code.  Anyone
// wanting to make use of these routines should get in touch with him, 
// prc@ccrma.stanford.edu.
// The following comments are taken from Perry Cook's SRC programme.
//
// 	The conversion is performed by sinc interpolation (ideal
// 	low-pass filter response).  One method uses sin() function
// 	calls from the floating point library, and the other method
// 	uses linearly interpolated table lookup as described in:
// 		J. O. Smith and P. Gossett, "A Flexible Sampling-Rate
// 		Conversion Method," Proc. of the IEEE Conference on
// 		Acoustics, Speech, and Signal Processing, San Diego,
// 		CA, March, 1984.  
//      The basic parameters select how wide
// 	the sinc function is for interpolation (how many neighboring
// 	samples are used), whether to use a table, and how finely the
// 	sinc is sampled in forming the table. 
// 
// 	The factor which specifies the conversion is a floating point
// 	number expressing the ratio of the old sampling rate to the new
// 	one.  The output sampling rate is left alone,
// 	so this operation as implemented causes pitch-shifting, and the
// 	factor is the ratio of the processed pitch to the original.
// 	The factor can also be viewed as the new sampling period,
// 	relative to the original sampling period.
//
//*****************************************************************************

// This controls the number of neighboring samples which are used to
// interpolate the new samples.  The processing time is linearly related to
// this width.

static const int WIDTH = 40;

// This controls whether a linearly interpolated lookup table is used for 
// sinc function calculation, or the sinc is calculated by floating point
// trig function calls. 

static const BOOL USE_TABLE = YES;

// This defines how finely the sinc function 
// is sampled for storage in the table.

static const int SAMPLES_PER_ZERO_CROSSING = 20;

static const double PI = 3.14159263;


// This is the size of the sample buffer.  It **MUST** be a power of two as we
// bit-mask it with one less than this to get legal array references into
// the sample buffer.

const int SIZE = 0x2000;


//*****************************************************************************

// static class member definitions.

double *srconv::sinc_table = make_sinc();
const int srconv::size_minus_one = SIZE - 1;
const int srconv::half_size = SIZE / 2;


//*****************************************************************************

// Constructor.  Fill the sample buffer with samples after allocating space
// for it.
// <params> is the line taken from the mappings file; <pathname> is the
// current path that the sounds are in; <line_num> is the line number of the
// mappings file that <params> is taken from.

srconv::srconv(char *params, char *pathname, int line_num)
               : sound(params, pathname, line_num)
{
    if (src != 1.0) {
	samplebuf = new short[SIZE];
	get_more_samples(samplebuf);
	get_more_samples(&samplebuf[half_size]);
    }
    else samplebuf = NULL;
}


//*****************************************************************************

// Deconstructor.

srconv::~srconv(void)
{
    delete[] samplebuf;
}


//*****************************************************************************
//
// Do the sampling-rate conversion of the buffer of samples.
// Called by soundlist::fillbuffers().
//
//*****************************************************************************
//
// In the words of the man that knows (i.e. here's Perry again): 
//
// True interpolation of sampled data
// signals involves ideal low-pass filtering in the frequency
// domain, which corresponds to convolution with a sinc function in
// the time domain (consult any text on Fourier transforms).
// 
// There are essentially two cases, one where the conversion
// factor is less than one, and the sinc table is used as is yielding a
// sound which is band limited to the 1/2 the new sampling rate
// (we don't want to create bandwidth where there was none).
// The other case is where the conversion factor is greater than
// one and we 'warp' the sinc table to make the final cutoff
// equal to the original sampling rate / 2.  Warping the sinc
// table is based on the similarity theorem of the time and
// frequency domain, stretching the time domain (sinc table)
// causes shrinking in the frequency domain (frequency response
// of ideal low-pass filter.
//
//*****************************************************************************


void srconv::srconvert(void)
{
    double one_over_src = 1.0;
    double temp1 = 0.0;
    double temp2 = 0.0;
    double temp3;
    static double time_now = 0.0;
    int i = 0;
    int j;
    int left_limit;
    int right_limit;
    static long next_stop = 0;
    long frames_needed = (long)ceil((double)sound::getRead_frames() * src);
    const long num_samples = Num_frames * Channels;

    // If this is the last call, set all the bytes in the sound's sample
    // buffer to zero so we don't play the end of the last buffer sent.
  
    if (frames_needed > frames_left) {
	frames_needed = frames_left;
	bzero((void *)samples, sound::getRead_frames() * 2 * sizeof(short));
    }
    next_stop += frames_needed;
    if (Channels == 1) {
	while (time_now < next_stop) {
	    temp1 = 0.0;
	    // leftmost neighboring sample used for interp.
	    left_limit = int(time_now - WIDTH + 1);
	    // rightmost leftmost neighboring sample used for interp.
	    right_limit = int(time_now + WIDTH); 
	    if (left_limit < 0) 
	      left_limit = 0;
	    if (right_limit > num_samples) 
	      right_limit = num_samples;
	    if (src < 1.0) {
		for (j = left_limit; j < right_limit; j++)
		  temp1 += get_sample(j, left_limit) * sinc(time_now - j);
		samples[i] = (short)temp1;
	    }
	    else {
		one_over_src = 1.0 / src;
		for (j = left_limit; j < right_limit; j++)
		  temp1 += get_sample(j, left_limit)
		    * sinc(one_over_src * (time_now - j));
		samples[i] = short(temp1 * one_over_src);
	    }
	    i++;
	    time_now += src;
	}
    }
    else {
	next_stop *= 2;
	while (time_now < next_stop) {
	    temp1 = 0.0;
	    temp2 = 0.0;
	    // leftmost neighboring sample used for interp.
	    left_limit = int(time_now - WIDTH + 1);
	    // rightmost leftmost neighboring sample used for interp.
	    right_limit = int(time_now + WIDTH); 
	    if (left_limit < 0) 
	      left_limit = 0;
	    if (right_limit > num_samples)
	      right_limit = num_samples;
	    if (src < 1.0) {
		for (j = left_limit; j < right_limit; j++) {
		    temp3 = sinc(time_now - j);
		    temp1 += get_sample(j * 2, left_limit) * temp3;
		    temp2 += get_sample(j * 2 + 1, left_limit) * temp3;
		}
	    }
	    else {
		for (j = left_limit; j < right_limit; j++) {
		    one_over_src = 1.0 / src;
		    temp3 = sinc(one_over_src * (time_now - j));
		    temp1 += get_sample(j * 2, left_limit) * temp3;
		    temp2 += get_sample(j * 2 + 1, left_limit) * temp3;
		}
	    }
	    samples[i * 2] = short(temp1 * one_over_src);
	    samples[i * 2 + 1] = short(temp2 * one_over_src);
	    i++;
	    time_now += src;
	}
    }
}

//*****************************************************************************

// Allocate, fill and return a pointer to the sinc table.

double *srconv::make_sinc(void)
{
    int i;
    double temp;
    double win_freq;
    double win;
    double *table;
    
    if (!USE_TABLE)
      return NULL;
    table = new double[WIDTH * SAMPLES_PER_ZERO_CROSSING];
    win_freq = PI / WIDTH / SAMPLES_PER_ZERO_CROSSING;
    table[0] = 1.0;
    for (i = 1; i < WIDTH * SAMPLES_PER_ZERO_CROSSING; i++) {
	temp = (double)i * PI / SAMPLES_PER_ZERO_CROSSING;
	table[i] = sin(temp) / temp;
	win = 0.5 + 0.5 * cos(win_freq * i);
	table[i] *= win;
    }
    return table;
}


//*****************************************************************************

// The sinc function.

double srconv::sinc(double x)
{
    double temp;
    
    if (USE_TABLE) 
      return t_sinc(x);
    else {
	if (x == 0.0)
	  return 1.0;
	else {
	    temp = PI * x;
	    return sin(temp) / temp;
	}
    }    
}

//*****************************************************************************

double srconv::t_sinc(double x)
{
    int low;
    double temp;
    double delta;
    
  if (fabs(x) > WIDTH - 1)
    return 0.0;
  else {
      temp = fabs(x) * (double)SAMPLES_PER_ZERO_CROSSING;
      // These are interpolation steps and can be ommited if desired 
      low = int(temp);    
      delta = temp - low;  
      return linear_interp(sinc_table[low], sinc_table[low + 1], delta);
      // For no interpolation case, return sinc_table[low]    
  }
}

//*****************************************************************************


inline double srconv::linear_interp(double first_number, double second_number,
				    double fraction)
{
    return (first_number + ((second_number - first_number) * fraction));
}

//*****************************************************************************


//*****************************************************************************
//*****************************************************************************
//*****************************************************************************
//
// The following is all my fault.  It's an attempt to interface transparently
// to the above code which was only moderately changed from Perry's NeXT     
// command line version.                                                     
//
//*****************************************************************************
//*****************************************************************************
//*****************************************************************************


//*****************************************************************************
//
// Perry's SRC code uses a NeXT SNDStruct, which treats the
// samples as a huge array, so you can reference any sample anywhere in
// the sound file just by giving an array reference.  On the SGI, we have
// to read a buffer's worth of samples at a time and when we've run out
// of samples, get another buffer.  When the SRC routine calls the sinc()
// function, we have a moving window of samples.  This is fine on the
// NeXT, but with our SGI buffered samples we need to be able to get a
// few samples beyond what is already in the buffer, but also go back to
// previous samples, i.e. not just refill the sample buffer once we need
// the sample one beyond our current buffer.  So we need a buffer that we
// initially fill with samples, then, when the left side of the window
// becomes the half-way point of the buffer, we fill the first half with
// new samples.  When the left side becomes the start of the buffer (by
// wrapping around using bit-masks), we fill the second half
// with new samples.  As we call this routine many times within the same
// window, we have to test to see which side we just filled before
// refilling.  Using this method we can just substitute an array
// reference in Perry's code with our get_sample() routine and everything
// else will be just fine (famous last words).
//
//*****************************************************************************


//*****************************************************************************

// If <left_limit> equals the half-way point and the last fill was
// the second half of the buffer, fill the first half of the buffer with
// new samples.  If <left_limit> equals 0, and the
// last fill was the first half of the buffer, fill the second half with
// new samples.

inline short srconv::get_sample(long aref, long left_limit)
{
    enum { FIRST, SECOND };
    int our_aref = aref & size_minus_one;
    int left = left_limit & size_minus_one;
    static int last_fill = SECOND;
    
    if (last_fill == SECOND && left == half_size) {
	get_more_samples(samplebuf);
	last_fill = FIRST;
    }
    else if (last_fill == FIRST && left == 0) {
	get_more_samples(&samplebuf[half_size]);
	last_fill = SECOND;
    }
    return samplebuf[our_aref];
}

  
//*****************************************************************************

// Get more samples from the sound file.  Store them in <start>.

inline void srconv::get_more_samples(short *start)
{
    long read = half_size / Channels;  // Frames to read. 
    long frames_read;
    
    read = frames_left > read ? read : frames_left;
    frames_read = fread((void *)start, sizeof(short) * Channels, read, Fp);
    frames_left -= frames_read;
    if (frames_read != read)
      cerr << "Warning: short read on sound ";
}
  
//*****************************************************************************

// EOF src.cc
