//*****************************************************************************
//
// File:          midi.cc
// Purpose:       Parse and store MIDI channel directions in the form
//                "midich = 12:4.5:1.3" where the colon separated
//                fields are the MIDI channel to read from, the
//                highest value to return and the lowest value to
//                return.  These last two fields are used to scale the
//                0-127 value sent by the MIDI volume message. 
// Author:        Michael Edwards
// Date:          4th March 1996
// Last revision: None
//
//*****************************************************************************

#include <ctype.h>
#include "mix.h"

// Max. number of chars in each number supplied to the <High> and
// <Low> parameters.

static const int WIDTH = 32;


// Declare Volumes array (static).

int midi::Volumes[16];


//*****************************************************************************

// Constructor: provide suitable defaults for the parameters.

midi::midi(void)
{
    Last_amp = 0.0;
    Channel = 0;
    High = 1.0;
    Low = 0.0;
    Data = NULL;
}


//*****************************************************************************

// Parse the argument to "midich=" which comes in the form of a single
// integer representing the MIDI channel, or in the form of
// <Channel>:<High>:<Low>.  Called in parameters::parameters()

void midi::parse(char *data)
{
    char hi[WIDTH] = "";   // Storage for getting <High> parameter.
    char lo[WIDTH] = "";   // Storage for getting <Low> parameter.
    char ch[3];            // Storage for getting <Channel> parameter.
			   // (Can only be 2 digits.)

    Data = data;
    split(ch, 3);
    Channel = atoi(ch);
    if (split(hi, WIDTH))
        High = atof(hi);
    if (split(lo, WIDTH))
        Low = atof(lo);
    if (Low >= High)
      cerr << "Warning:  Low MIDI value greater than or equal to High:"
	      "\nChannel = " << Channel << ", High = " << High << ", Low = "
	   << Low << '\n';
    if (Channel < 0 || Channel > 16)
      error("midi::parse", "MIDI channel can only be between 0 and 16:\n"
	    "You specified %d in \"%s\"", Channel, data);
	    
}


//*****************************************************************************

// Get the next argument in the colon delimited list.  Do all syntax
// error checking necessary.  Increments the <Data> pointer over the
// processed characters.
// Called by midi::parse()

BOOL midi::split(char *holder, int limit)
{
    int i = 0;
    char *ptr = Data;

    // Skip whitespace.
    while (isspace(*ptr))
        ptr++;

    while (i < limit) {

	// Get the number.
	if (isdigit(*ptr) || *ptr == '.')
	    holder[i++] = *ptr++;
	
	// If there's space now, it should only be before the colon or
	// end of string, so signal an error if this isn't the case.
	else {
	    while (isspace(*ptr))
	        ptr++;
	      
	    // Stop at the colon delimiter.
	    if (*ptr == ':') {
		ptr++;
		break;
	    }
	    
	    // Stop at the end of the string.
	    else if (*ptr == '\0')
	      break;
	    else error("midi::split", "Syntax error: %s", Data);
	}
    }

    // User could have given a number longer than <limit> chars so
    // signal error if this is the case--it won't be parsed correctly.
    if (i == limit)
	error("midi::split", "Number too long: %s", Data);

    // Update the Data pointer to skip the processed characters.
    Data = ptr;

    // Return YES if we've processed some characters, otherwise NO so
    // we don't overwrite the default values supplied in the
    // constructor. 
    if (i > 0)
        return YES;
    else
        return NO;
}


//*****************************************************************************

// EOF midi.cc

