//*****************************************************************************
//
// File:     mix.cc
// Purpose:  The main mixing routine, including interpreting the mappings file,
//           forking the MIDI and audio processes, and catching keys and
//           playing the associated sound file.
//           Key-to-soundfile mapping is done in a text file.  Specifying
//           "path" at the start of a line will automatically define the
//           directory in which the sounds are to be found (can be done at any
//           point within the file).  Comments are inserted by starting
//           a new line with a ";" and all empty lines will be ignored.
//           Each sound can also be assigned a MIDI channel.  This is the
//           channel that will be read when scaling the amplitude.  A MIDI
//           fader box sends velocity values that are used for scaling 
//           purposes.  If no amplitude control is needed, set the channel to
//           0 (the default).
//
//           E.g.
//           path /usr/lib/sounds/prosonus/ambience/ 
//           amp = .6, key = q, snd = rain.snd, midich = 5 
//           key = w, snd = crickets.aiff, midich = 6 
//           ; This is a comment.
//           key = e, snd = forest.aiff, midich = 7 
//           key = r, snd = traffic.snd, midich = 8 
//           path /usr/lib/sounds/prosonus/instr/ 
//           key = t, snd = cello_pizz.C4.snd, midich = 10 
//
//           See parameters.cc for the full syntax conventions.
// Author:   Michael Edwards
// Date:     October 3rd 1995
// Last revision: 21st November 1995.
//
//*****************************************************************************

#include "mix.h"
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <math.h>

//*****************************************************************************

// Function prototypes: 

static void catch_control_c(int signal);


//*****************************************************************************

// Globals: 

static BOOL Caught_sigint;      // Whether user pressed Control-C or not. 


//*****************************************************************************

// Constructor.  Interpret the mappings file, inititialize the
// <Mappings> array, and prepare the sounds for playing.  Set up the
// terminal to catch characters without echoing.  <file> is the path
// (relative or full) of the mappings file. 
 

mix::mix(char *file)
{
    FILE *fp;
    char path[MAXPATH] = "";    // The directory the sounds are in. 
    char line[MAXLINE];         // One line of text from the file. 
    int i;
    int aref;                   // The array reference of the mappings array.
    sound *snd;

    
    // Initialize all the sound pointers to NULL so we can test for the
    // presence of a mapped file. 
    for (i = 0; i < MAXSOUNDS; i++)
        Mappings[i] = NULL;

    // Initialise the midi volumes to max. (127).
    midi::initVolumes();
    
    // Interpret the mappings file. 
    if ((fp = fopen(file, "r")) == NULL)
        error("mix::mix", "Can't open \"%s\".", file);
    for (i = 1; fgets(line, MAXLINE, fp) && line != NULL; i++) {
        stringtrim(line);       // Get rid of spaces at either end. 
        
        // Ignore comments and empty lines.
        if (*line == ';' || *line == '\n' || *line == '\0')
            continue;
        // If line starts with "path" get the new path and store it. 
        if (!strncmp(line, "path", 4)) {
            sscanf(line, "path %s", path);
            continue;
        }

        // else the line contains mapping info. 
        else
            snd = new sound(line, path, i);
        
        // Get the array reference of where we're going to store this sound
        // in the mappings array. 
        aref = snd->getKey() - 'A';

        // If we haven't already mapped to this key, do so. 
        if (Mappings[aref] == NULL)
            Mappings[aref] = snd;
        else
            error("mix::mix", "Two sounds mapped to the same key: \n\r"
                  "Sound: \"%s\", Key: %c",
                  snd->getPath(), snd->getKey());
    }

    // Print all kinds of useful sound info to the terminal.  Prints the keys
    // in alphabetical order. 
    for (i = 0; i < MAXSOUNDS; i++)
        if (Mappings[i] != NULL)
            Mappings[i]->printStats();
    fclose(fp);
}


//*****************************************************************************

// Close files, free memory etc., and do general clean up. 

mix::~mix(void)
{
    int i;
    sound *snd;

    // Wait for the semaphore. 
    GETSEMA();

    // Stop the sounds and free all the sound resoures. 
    for (i = 0; i < MAXSOUNDS; i++) {
        snd = Mappings[i];
        if (snd != NULL) {
            
            // If the sound is playing, stop it first before closing the
            // sound file. 
            if (snd->getPlaying() == YES)
                snd->setPlaying(NO);
            delete snd;
        }
    }
    
    // Free the temporary buffer for the sound class--this is a static member
    // so there's only one. 
    sound::freeTempbuf();
}

//*****************************************************************************

// Once the sounds have been mapped, play the file associated with the key
// pressed.  Also catch Control-C and make sure system resources are freed
// etc.  We use semaphores in this function because this is where both the
// MIDI and audio processes are forked.  The MIDI process and
// the audio process both set and access globals so we don't want to
// do anything with these globals 
// unless it's safe to do so.  Using a global semaphore allows us to test
// whether it's safe to perform certain operations or not.  We wait for the
// semaphore before doing these operations and release it to the other
// processes afterwards. <timestampfile> is the file to which timing
// information is written.  Nothing is written if this is NULL; 

void mix::run(char *timestampfile)
{
    int c;      // The character caught. 
    int aref;   // The array reference for the Mappings array. 
    sound *snd; // The sound accessed in the Mappings array. 

#ifndef DEBUG
#ifdef SGI
    // The process ID of the MIDI process.  Used to kill it when we're done. 
    pid_t midi_pid;

    // The process ID of the audio process. Sim. to midi_pid. 
    pid_t audio_process_pid;
    int i;
#endif  // SGI 

#ifdef NEXT
    cthread_t midi_pid;
    cthread_t audio_process_pid;
#endif  // NEXT 
#endif  // DEBUG 

    // The sound list we're going to pass to the audio process. 
    soundlist sndlist;
    FILE *fp = NULL;    // Timestamp stream. 

#ifdef SGI

    // Shared Arena and semaphore stuff: 
    char tempfile[] = "/tmp/mix_XXXXXX"; // A template for the temp file. 
    usptr_t *arena;                      // The shared arena. 

    mktemp(tempfile);
    arena = usinit(tempfile);            // Initialize the shared arena. 
    unlink(tempfile);
    Semaphore = usnewsema(arena, 0);     // Allocate a new semaphore. 
#endif  // SGI 

#ifdef NEXT
    Semaphore = mutex_alloc();           // Allocate a new semaphore. 
#endif  // NEXT 

    // Catch Control-C 
    signal(SIGINT, (void (*) (int))catch_control_c);

    // Open the timestamp file if we need it: 
    if (timestampfile != NULL)
        if ((fp = fopen(timestampfile, "w")) == NULL)
            error("mix::run", "Can't open \"%s\" for writing.", timestampfile);
    
    // Fork the MIDI and audio processes and store the PID's so we can kill
    // them later. 
     
#ifndef DEBUG
#ifdef SGI
    if ((midi_pid = sproc(accept_volumes, PR_SALL)) == -1)
        error("mix::run", "Couldn't start MIDI process.");
    if ((audio_process_pid = sproc(audio_process, PR_SALL, (void *)&sndlist))
        == -1)
        error("mix::run", "Couldn't start audio process.");
#endif  // SGI 

#ifdef NEXT
    midi_pid = cthread_fork((cthread_fn_t) accept_volumes,
                            (any_t) 0);
    audio_process_pid = cthread_fork((cthread_fn_t) audio_process,
                                     (any_t) & sndlist);
    cthread_detach(midi_pid);
    cthread_detach(audio_process_pid);
#endif  // NEXT 
#endif  // DEBUG 

    // Release the semaphore to the forked processes. 
    RELEASEMA();
    if (!Silent) {
        printf("\n\r");
        if (TestMidi)
          printf("MIDI:   1   2   3   4   5   6   7   8   9  10"
                 "  11  12  13  14  15  16    ");
	printf("Underflows  Limits      Sounds\n\r");
        fflush(stdout);
    }
    Caught_sigint = NO;

    //********************** Main key catching loop: ***********************

    while (!Caught_sigint) {
#ifdef DEBUG
        c = getchar();  // Get the keyboard character--buffered.
#else
        c = getch();    // Get the keyboard character--unbuffered.
#endif

        // Get the array reference. 
        aref = c - 'A';

        // Only process upper- and lower-case alphabet input, as well
        // as those characters inbetween in the ASCII table ([ \  ]  ^  _  `)
        if (aref >= 0 && aref < MAXSOUNDS) {

            // Wait for the semaphore. 
            GETSEMA();

            // Get the sound mapped to this key. 
            snd = Mappings[aref];
            
            // If there's a sound mapped to this key and it's not playing,
            // play it. 
            if (snd != NULL) {
                if (snd->getPlaying() == NO) {

                    // Let the sound know it's playing. 
                    snd->setPlaying(YES);
                    snd->JustStarted = YES;

                    // Add the sound to the sound list. 
                    sndlist.add(snd);

                    // Time stamp, if requested. 
                    timestamp(fp, snd, YES);
                }
                else {
                    snd->JustStopped = YES;
                    timestamp(fp, snd, NO);
                    if (snd->getRestart()) 
                        timestamp(fp, snd, YES);
                }

                /*
                // If <Restart> was set, reset the sound (rewind to beginning
                // of samples) and tell it it's playing again. 
                else if (snd->getRestart()) {
                    snd->reset();
                    snd->setPlaying(YES);
                    
                    // This key click stopped and restarted so we need two
                    // timestamps. 
                    timestamp(fp, snd, NO);
                    timestamp(fp, snd, YES);
                }

                // else stop the sound.
                else {

                    // soundlist::tidy() will remove the sound from the
                    // soundlist if snd->Playing is NO. 
                    snd->setPlaying(NO);
                    timestamp(fp, snd, NO);
                }
                */
            }

#ifdef DEBUG
            audio_process((void *)&sndlist);
#endif

            // Release the semaphore to the other processes. 
            RELEASEMA();
        }
        // User pressed a number key, so interpret it's functionality. 
        else if (isdigit(c))
            get_input(c);

        // Stop all the sounds if "!" is pressed. 
        else if (c == '!') {
            GETSEMA();
            sndlist.stopall();
            RELEASEMA();
        }
    }
    
    //******************* End of main key catching loop: *********************

    // General cleanup.  
    // Close the time stamp file, if we opened one. 
    if (fp != NULL)
        fclose(fp);

#ifndef DEBUG
#ifdef SGI
    // Kill the MIDI process. 
    if ((i = kill((pid_t) midi_pid, SIGKILL)) == -1)
        fprintf(stderr, "Couldn't kill MIDI process.\n\r");

    // Kill the audio process. 
    if ((i = kill((pid_t) audio_process_pid, SIGKILL)) == -1)
        fprintf(stderr, "Couldn't kill audio process.\n\r");
    // Free sound resources.
    close_ports();
#endif // SGI

#ifdef NEXT
    // Kill the MIDI process.
    thread_suspend(cthread_thread(midi_pid));
    cthread_abort(midi_pid);

    // Kill the audio process. 
    thread_suspend(cthread_thread(audio_process_pid));
    cthread_abort(audio_process_pid);
    
    // Free the semaphore.
    mutex_free(Semaphore);
#endif // NEXT
#endif // DEBUG 

    fprintf(stderr, "\n\rBye.\r\n");
}


//*****************************************************************************

// The user pressed a number key which indicates changing some parameter on
// the fly, so interpret the functionality here. <key> is the key that was
// pressed. 
 

void mix::get_input(int /* key */ )
{
#ifdef LATER
    char buf[32];  // Storage for user input. 

    // Step out of curses mode so we can echo chars again. 
    endwin();

    // Stop printing the underflow errors etc. 
    OK_to_print = NO;
    switch (key) {
    case '1':
        printf("Type SRC: ");
        fflush(stdout);
        gets(buf);
        break;
    case '2':
        printf("Play the next sound backwards? (y or n) (<cr> = no) ");
        fflush(stdout);
        gets(buf);
        break;
    default:
        printf("*** Key's functionality not implemented. ***");
        fflush(stdout);

#ifdef SGI
        sginap(80);
#endif // SGI

#ifdef NEXT
        sleep(1);
#endif // NEXT

    }

    // Now we can print underflow errors again. 
    OK_to_print = YES;

    // Get back into curses mode. 
    refresh();
#endif  // LATER 
}

//*****************************************************************************

// Record which key was pressed, what sound file it triggered and at what
// time it was triggered.  Write the information into the file pointed to by
// <fp>.  Also record whether the key press started or stopped a sound. Do
// nothing if no file was specified to write to. <fp> is the stream to write
// to, <snd> is the sound object and <start> tells us
// whether this was a key press that started a sound or stopped it. Called by
// mix::run(). 
 
void mix::timestamp(FILE *fp, sound *snd, BOOL start)
{
    if (fp == NULL)
         return;
    else {
        struct timeval tv;
        double thistime;        // The time of this key press. 
        char *msg = (start == YES ? "started at" : "**stopped** at");

        // The time of the first key press--only set once. 
        static double starttime;

        // Flag so that we don't set starttime more than once. 
        static BOOL got_start = NO;

        gettimeofday(&tv, NULL);
        thistime = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);
        
        // If this was the first key press, get the time and store it for use
        // in every subsequent time stamp. 
        if (!got_start) {
            starttime = thistime;
            got_start = YES;
        }
        thistime -= starttime;
        fprintf(fp, "Key %c (%s) %s %.4f secs.\n",
                snd->getKey(), snd->getPath(), msg, thistime);
    }
}


//*****************************************************************************

char* GetTime(BOOL first_time)
{
    static char buf[256] = "0:00.00";

    if (!first_time)
        return buf;

    struct timeval tv;
    double thistime;        // The time of this key press. 

    // The time of the first key press--only set once. 
    static double starttime = 0.0;

    gettimeofday(&tv, NULL);
    thistime = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);
        
    // If this was the first key press, get the time and store it for use
    // in every subsequent time stamp. 
    if (starttime == 0.0)
        starttime = thistime;
    thistime -= starttime;
    
    double mins = floor(thistime / 60.0);
    double secs = thistime - (mins * 60.0);
    sprintf(buf, "%d:%02.2f", int(mins), secs);
    return buf;
}


//*****************************************************************************

// Set <Caught_sigint> when user presses Control-C so that the main key
// catching loop breaks. Called by mix::run() via sigset(). 

static void catch_control_c(int signal)
{
    if (signal == SIGINT) {
        Caught_sigint = YES;

#ifdef NEXT
        // Curses on the NeXT doesn't seem to catch Control-C with a
        // call to getch(), so ask the user to hit another key to
        // exit. 
        OK_to_print = NO;
        fprintf(stderr, "\nPress any key to exit.\n");
        fflush(stdout);
#endif // NEXT

    }
}


//*****************************************************************************

// EOF mix.cc 
 
