//*****************************************************************************
//
// File:          sndfile.cc
// Purpose:       Open an AIFF or SND sound file and read it's header,
//                storing sound info in appropriate class members.
//                Reads the different sound formats transparently, so
//                all the user has to do is declare the class with the
//                pathname as argument to the constructor and
//                everything will be taken care of, including printing
//                errors if the file isn't a sound etc.
// Author:        Michael Edwards
// Date:          December 19th 1996
// Last revision: None
//
//*****************************************************************************

#include <string.h>
#include "mix.h"


//*****************************************************************************

// The number of characters to read from the header.  Should be easily enough
// to get all the info we need. 

static const int HDRSIZE = 512;


//*****************************************************************************

// Constructor.  Needs the full path of the sound file (<path>) which is
// immediately stored in the class member <Path>. Open the file, allocate
// <Hdrbuf> and read in <HDRSIZE> bytes.  Test to see whether the file is
// AIFF or SND and process accordingly.  Free Hdrbuf at the end as we won't
// be needing header info anymore. 

sndfile::sndfile(const char *path)
{
    Fp = fopen(path, "r");
    if (Fp == NULL)
	error("sndfile::sndfile", "Can't open %s.", path);
    Hdrbuf = new char[HDRSIZE];

    fread(Hdrbuf, 1, HDRSIZE, Fp);
    if (!strncmp(Hdrbuf, "FORM", 4))	   // It's an AIFF file. 
	read_aiff(path);
    else if (!strncmp(Hdrbuf, ".snd", 4))  // It's an SND file. 
	read_snd(path);
    else
	error("sndfile::sndfile",
	      "Unsupported sound file type or file is not a sound file:\n"
	      "\"%s\"",
	      path);

    // Don't need the header info anymore. 
    delete[] Hdrbuf;
}


//*****************************************************************************

// Destructor.  Simply close the file pointer. 

sndfile::~sndfile(void)
{
    fclose(Fp);
}


//****************************************************************************
//
// Byte positions for ".aiff" type sound files:
// (Taken from Bill Schottsteadt's headers.c)
//
// 0: "FORM"
// 4: size (bytes)
// 8: "AIFF" or "AIFC" -- the latter includes compressed formats.
//
// Thereafter the file is organized into "chunks", each chunk being
// a 4-byte identifer followed by an int (4-bytes) giving the chunk size
// not including the 8-byte header.  AIFF data is signed.  If the chunk
// size is odd, an extra (unaccounted-for) null byte is added at the end.
//
// The chunks we want are "COMM" and "SSND".
//
// COMM: 0: chans
//       2: frames
//       6: bits per sample
//       8: srate as 80-bit IEEE float
// then if AIFC (not AIFF), 4 bytes giving compression id ("NONE"=not
// compressed) followed by Pascal string giving long name of compression
// type.
// SSND: 0: data location (offset within SSND chunk).
//
//*****************************************************************************


// Read an AIFF header. 

void sndfile::read_aiff(const char *path)
{
    // The sound file format.  Should be 16 (which is 16-bit linear). 
    int format;
    // Counts the number of bytes processed. 
    int count;
    // We're going to be incrementing this pointer and we don't want
    // to lose the <Hdrbuf> pointer because we need to free it (all). 
    char *ptr = Hdrbuf;

    // Make sure this is an AIFF file and not an AIFC compressed file. 
    if (strncmp(ptr + 8, "AIFF", 4) != 0)
	error("sndfile::read_aiff",
	      "Not supporting AIFF compressed format.\n%s.",
	      path);
    ptr += 12;			// Skip the first block "FORMxxxxAIFF" 
    count = 12;			// Ditto 
    
     // Skip along to the COMM block--we should already be there, but best to
     // be safe. 

    while (strncmp(ptr, "COMM", 4)) {
	count += 2;
	ptr += 2;
    }

    // Get the data from the correct byte positions as detailed above. 
    Channels = get_short(ptr + 8);
    Num_frames = get_int(ptr + 10);

    // format of 16 is 16-bit linear (same as format 3 in ".snd" files). 
    format = get_short(ptr + 14);
    if (format != 16)
	error("sndfile::read_aiff", "All sound files must be 16-bit linear.\n"
	      "Offending file:  %s", path);

    // Convert that horrible sampling rate double format. 
    Srate = (int)ieee_80_to_double(ptr + 16);
    
    // Skip along to the SSND block to get the <Data_location>. According to
    // the specification, we should be able to get the chunk size and skip
    // that many bytes to read the next chunk header until we get to the SSND
    // block.  However, when we have header sizes above 200 bytes this
    // doesn't seem to work (can't find out why).  So we have to skip two
    // bytes at a time and see if we got to the SSND block the slow way. 
    while (strncmp(ptr, "SSND", 4)) {
	count += 2;
	
	// Don't generate pointer errors!  If we haven't got there by now,
	// something is wrong. 
	if (count > HDRSIZE) 
	  error("sndfile::read_aiff", "Error reading header: %s", path);
	ptr += 2;
    }
    Data_location = get_int(ptr + 8) + count + 16;
}


//****************************************************************************
//
// Byte positions for ".snd" type sound files:
// (Taken from Bill Schottsteadt's headers.c)
//
//   0:  ".snd"
//   4:  data_location (bytes) (not necessarily word aligned on Sun)
//   8:  data_size (bytes) -- sometimes incorrect ("advisory")
//   12: data format indicator -- see below
//   16: srate (int)
//   20: chans
//   24: comment start
//
// Formats are:
//
// 0 unspecified, 1 mulaw_8, 2 linear_8, 3 linear_16, 4 linear_24,
// 5 linear_32, 6 float, 7 double, 8 indirect, 9 nested, 10 dsp_core,
// 11 dsp_data_8, 12 dsp_data_16, 13 dsp_data_24, 14 dsp_data_32,
// 16 display, 17 mulaw_squelch, 18 emphasized, 19 compressed,
// 20 compressed_emphasized, 21 dsp_commands, 22 dsp_commands_samples,
// 23 adpcm_g721, 24 adpcm_g722, 25 adpcm_g723, 26 adpcm_g723_5,
// 27 alaw_8, 28 aes, 29 delat_mulaw_8.
//
//*****************************************************************************


// Read an SND header.  This is a lot less hassle than AIFF. 

void sndfile::read_snd(const char *path)
{
    int format;

    Data_location = get_int(Hdrbuf + 4);
    format = get_int(Hdrbuf + 12);
    if (format != 3)
	error("sndfile::read_snd", "All sound files must be 16-bit linear.\n"
	      "Offending file:  \"%s\"", path);
    Srate = get_int(Hdrbuf + 16);
    Channels = get_int(Hdrbuf + 20);
    
     // Data size is in bytes so do appropriate scaling to get the number of
     // frames. 
    Num_frames = get_int(Hdrbuf + 8) / (sizeof(short) * Channels);
}


//*****************************************************************************

// Get a 4-byte integer from the character pointer <start>.  Get it right
// whether this is little or big endian. 

inline int sndfile::get_int(char *start)
{
    unsigned char c1 = *start;
    unsigned char c2 = *(start + 1);
    unsigned char c3 = *(start + 2);
    unsigned char c4 = *(start + 3);

//#ifdef LITTLE_ENDIAN
//    return (((int)c4 << 24) | ((int)c3 << 16) | ((int)c2 << 8) | (int)c1);
//#else
    return (((int)c1 << 24) | ((int)c2 << 16) | ((int)c3 << 8) | (int)c4);
//#endif // LITTLE_ENDIAN
}


//*****************************************************************************

// Same as get_int, only return a 2-byte short. 

inline short sndfile::get_short(char *start)
{
    unsigned char c1 = *start;
    unsigned char c2 = *(start + 1);

//#ifdef LITTLE_ENDIAN
//    return (((short)c2 << 8) | (short)c1);
//#else
    return (((short)c1 << 8) | (short)c2);
//#endif // LITTLE_ENDIAN
}


//*****************************************************************************

// Following two functions stolen wholesale from Bill Schottsteadt's
// headers.c.  Called by sndfile::read_aiff(). 

double sndfile::ieee_80_to_double(char *p)
{
    char sign;
    short exp = 0;
    unsigned long mant1 = 0;
    unsigned long mant0 = 0;
    double val;

    exp = *p++;
    exp <<= 8;
    exp |= *p++;
    sign = (exp & 0x8000) ? 1 : 0;
    exp &= 0x7FFF;
    mant1 = *p++;
    mant1 <<= 8;
    mant1 |= *p++;
    mant1 <<= 8;
    mant1 |= *p++;
    mant1 <<= 8;
    mant1 |= *p++;
    mant0 = *p++;
    mant0 <<= 8;
    mant0 |= *p++;
    mant0 <<= 8;
    mant0 |= *p++;
    mant0 <<= 8;
    mant0 |= *p++;
    if (mant1 == 0 && mant0 == 0 && exp == 0 && sign == 0)
	return 0.0;
    else {
	val = myUlongToDouble(mant0) * pow(2.0, -63.0);
	val += myUlongToDouble(mant1) * pow(2.0, -31.0);
	val *= pow(2.0, ((double)exp) - 16383.0);
	return sign ? -val : val;
    }
}


//*****************************************************************************

// Called by sndfile::ieee_80_to_double(). 

inline double sndfile::myUlongToDouble(unsigned long ul)
{
    const unsigned long ULPOW2TO31 = 0x80000000L;
    const double DPOW2TO31 = 2147483648.0;
    double val;

    if (ul & ULPOW2TO31)
	val = DPOW2TO31 + (ul & (~ULPOW2TO31));
    else
	val = ul;
    return val;
}


//*****************************************************************************

// EOF sndfile.cc 
 
