//*****************************************************************************
//
// File:          sound.cc
// Purpose:       Do all the necessary calculations and file handling
//                routines before playing a sound.
// Author:        Michael Edwards
// Date:          30th September 1995
// Last revision: 9th November 1995, 20th November 1995
//
//*****************************************************************************

#include <errno.h>
#include <fcntl.h>
#ifdef SGI
#include <bstring.h>
#endif
#ifdef NEXT
#include <string.h>
#endif
#include <iostream.h>
#include "mix.h"


#ifdef SGI
// Number of seconds of sound read each time by AFreadframes(). This should
// yield an integer when 1.0 is divided by it.   Only change this if
// you're running on an SGI.
static const double SECS_PER_BUF = 0.025;
#endif // SGI

#ifdef NEXT
// This value was arrived at empirically--change at your peril.
static const int READ_FRAMES = 2050;
#endif // NEXT

// Declarations for static data members.
long sound::Read_frames = 0L;
long sound::Sampling_rate = 0L;
float* sound::MonoRamp = NULL;
float* sound::StereoRamp = NULL;


//*****************************************************************************

// Constructor.                                                            
// Prepare a sound file for playing.                                       
// <params> is the line taken from the mappings file; <pathname> is the    
// current path that the sounds are in; <line_num> is the line number of the
// mappings file that <params> is taken from. 
// Call the constructors for the superclasses here.

sound::sound(char *params, char *pathname, int line_num) 
             : parameters(params, pathname, line_num), sndfile(Path)
{
    // Only set <Read_frames> once. 
    if (!Read_frames) {

#ifdef SGI
        Read_frames = (int)(Srate + 1) / (int)(1.0 / SECS_PER_BUF);
#endif // SGI

#ifdef NEXT
        Read_frames = READ_FRAMES;
#endif // NEXT

	if (MonoRamp == NULL) {
	  MonoRamp = new float[Read_frames];
	  MakeRampDownBuf(MonoRamp, Read_frames);
	}
	
	if (StereoRamp == NULL) {
	  StereoRamp = new float[Read_frames * 2];
	  MakeRampDownBuf(StereoRamp, Read_frames * 2);
	}
	
        // Allocate memory for <Tempbuf>.
        initTempbuf(Read_frames * 2);
    }
    Duration = (double)Num_frames / Srate;
    
    // If we have a skip, that's where we start from no matter whether
    // there's reverse or not. 
    if (Skip)
        Start_frame = (long)(Skip * Srate);
    else if (Reverse)
        Start_frame = (Num_frames - Read_frames) - 1;
    else
        Start_frame = 0;
    Start_byte = Data_location + (Start_frame * sizeof(short) * Channels);
    
    // If <Reverse> and <Skip> are YES, we start at <Start_frame> and read
    // backwards so the number of frames is equal to <Start_frame>. 
    if (Reverse && Skip)
        Num_frames = Start_frame;
    else if (Skip)
        Num_frames -= Start_frame;
    if (Num_frames <= 0)
        error("sound::sound",
            "Empty sound file or skip longer than sound's duration: \n\r%s:"
              " duration %f, requested Skip %f",
              Path, Duration, Skip);
    
    // Make the buffer large enough to hold SECS_PER_BUF secs of audio
    // frames. We added one to sampling_rate (above) before we divide in
    // order to correctly handle the 11025 case. The sound buffer is always
    // enough for SECS_PER_BUF of **stereo** samples, because mono
    // samples need to be converted to stereo before mixing. 
    Samples = new short[Read_frames * 2];

    // Go to sample number Start_frame (Start_frame or default 0): 
    reset();

#ifdef NEXT
    // Set sampling rate the first time.  This is done separately on
    // the SGI by setoutputrate().
    if (Sampling_rate == 0)
        Sampling_rate = Srate;
#endif // NEXT

    // Make sure all files are at the same sampling rate.
    if (Sampling_rate != 0 && Srate != Sampling_rate)
      error("sound::setoutputrate",
            "All sounds must be at the same sampling rate.\n"
            "Offending file: \"%s\"", Path);

#ifdef SGI
    setoutputrate((long)Srate);
#endif // SGI

    JustStarted = NO;
    JustStopped = NO;
}


//*****************************************************************************

// Destructor. 

sound::~sound(void)
{
    delete[] Samples;
}


//*****************************************************************************

// Allocate the temporary buffer where samples are processed (mono
// to stereo, reverse etc.).  This has to be done here as <Tempbuf>
// is a static member variable, so needs to be initialized only once.
// We only need one <Tempbuf> per class because sounds are converted
// one at a time, never in parallel.  Called in sound::sound().

// Declare <tempbuf> first.
short *sound::Tempbuf = NULL;

void sound::initTempbuf(long siz) 
{
  Tempbuf = new short[siz];
}


//*****************************************************************************

void sound::MakeRampDownBuf(float* buf, long size)
{
    const float decrement = 1.0f / float(size - 1);
    float val = 1.0f;
    float* b = buf;

    for (int i = 0; i < size; ++i) {
        *b = val;
        ++b;
        val -= decrement;
    }
}


//*****************************************************************************

// Free memory allocated for <Tempbuf>.  Called in mix::~mix().

void sound::freeTempbuf(void)
{
    delete[] Tempbuf;
    Tempbuf = NULL;
    delete[] MonoRamp;
    MonoRamp = NULL;
    delete[] StereoRamp;
    StereoRamp = NULL;
}


//*****************************************************************************

// Seek to the beginning frame (or the position specified by <Skip>), reset
// the number of frames left to what it was originally and tell the sound
// it's not playing.  Seek to the end of the sound minus one buffer's worth
// of samples if Reverse is set and Skip = 0. Called by sound::sound(),
// soundlist::tidy() and soundlist::fillbuffers(). 
//
// The reason this function is used twice in the soundlist class functions is
// that there are two situations in which we have to reset the sound back to
// it's initial state.  One, when the sound dies a natural death, i.e. we
// play all it's samples (the fillbuffers() case); and two, when the user
// kills the sound by pressing it's associated key while it is playing (the
// tidy() case). 

void sound::reset(void)
{
    if (fseek(Fp, Start_byte, SEEK_SET) != 0) {
         fprintf(stderr, "Couldn't reset %s back to the start point\n\r",
                  Path);
        fflush(stdout);
    }

    // Reset the number of frames left for the next time we play. 
    Frames_left = Num_frames;
    
    // Tell the sound it's not playing anymore so that the next key press
    // will start it again at the beginning. 
    Playing = NO;
}


//*****************************************************************************

// Print the parameter values of a Parameters structure, as well as
// information about the sound. Called by mix::mix(). 

void sound::printStats(void)
{
    // Long printout version.
    if (Info)
         printf("Key = %c, Sound = %s, \n\r\t[duration %.4f, %d channels, "
                "%.2f KHz sampling rate],\n\r\tMIDI channel = %d (Max %.4f, "
                "Min %.4f), Skip = %.4f, \n\r\tAmplitude = %.4f, %s %s\n\r",
                Key, Path, Duration, Channels,
                Srate / 1000.0, Midich.getChannel(), Midich.getHigh(),
                Midich.getLow(), Skip, Amp, Reverse ? "*** REVERSED ***" : "",
                Restart ? "*** RESTART ***" : "");

    // Short version.
    else {
        char buf[32];

        filename(Path, buf);
        printf("%c   %2d   %-34s   %s\n\r", Key, Midich.getChannel(), buf,
               Restart ? "*** RESTART ***" : "");
    }
    fflush(stdout);
}


//*****************************************************************************

// Convert a buffer of samples from mono to stereo simply by copying the
// mono samples to <Tempbuf>, then copying each one twice back in to the
// <samples> buffer of the sound object. Called by soundlist::fillbuffers(). 

void sound::mono_to_stereo(void)
{
    register long i;
    register short *current_sample = Tempbuf;
    register short *samps = Samples;

    bcopy((void *)Samples, (void *)Tempbuf, sizeof(short) *
          Read_frames);

    // We need to loop up to Read_frames because this is the half way point
    // in a stereo buffer. 
    for (i = 0; i < Read_frames; ++i) {
        *samps++ = *current_sample;
        *samps++ = *current_sample++;
    }
}


//*****************************************************************************

// Scale the samples in a buffer by the amount specified as a global scaler
// in the key-mappings file, and by the MIDI volume from the channel
// associated with this sound.  Call this routine before
// mono_to_stereo() to avoid a lot of unnecessary multiplies.
// Called by soundlist::fillbuffers(). 
// We interpolate between the last
// amplitude value sent by the MIDI fader and the new amplitude so that a
// smooth change of volume is effected rather than just a step from one
// buffer of samples to the next. 
 

void sound::scale(void)
{
    // The product of the global scaler and the MIDI volume. 
    double new_midi;
    register int i;
    short *samps = Samples;
    const double num = double(Read_frames * Channels);
    double inc;    //The increment for interpolation.
    register double last_midi = Midich.getLast_amp();

    // No need to scale the samples if no MIDI channel was assigned and the
    // global amplitude given (by default, perhaps) was 1.0. 
    if (Midich.getChannel() == 0 && Amp == 1.0)
        return;
    new_midi = Midich.get_dvalue();

    // Get the right increment for interpolation        .
    inc = (new_midi - last_midi) / num;

    // If this is the first buffer we don't want to ramp from the last
    // amplitude we were at.
    if (JustStarted) {
      inc = 0.0;
      last_midi = new_midi;
      JustStarted = NO;
    }

    // Scale 'em all:
    for (i = 0; i < num; i++) {

        // Interpolate from last amp to the new one.
        last_midi += inc;
        *samps = short(double(*samps) * last_midi * Amp);
        ++samps;
    }
}



//*****************************************************************************

// Reverse a buffer of samples so that playing a sound backwards is possible.
// Call this routine before mono_to_stereo(). Called by
// soundlist::fillbuffers(). 
//
//***************************** Warning ***************************************
//
// One side effect of this process is that channels will be switched.  We
// simply copy the last sample into the first sample and go on like this
// incrementing and decrementing, so if the sound is stereo, the last sample
// will be the right channel which when copied becomes the first sample,
// which is the left channel. 

void sound::reverse_buffer(void)
{
    short *tempbufptr = Tempbuf;
    short *sampleptr = Samples;
    int i = Read_frames * Channels;

    // Copy the samples into <Tempbuf>. 
    bcopy((void *)Samples, (void *)Tempbuf,
          sizeof(short) * i);

    // Go to the end of the sample buffer. 
    tempbufptr += (Read_frames * Channels) - 1;
    while (i-- > 0)
        *sampleptr++ = *tempbufptr--;
}


//*****************************************************************************

long sound::readsamples(long num)
{
    long read = readsamples_aux(num, YES);

    if (read == 0)
        return 0;

    long n = Read_frames * Channels;

    if (JustStopped) {
        short* smps = Samples;
        float* ramp = (Channels == 1 ? &MonoRamp[0] : &StereoRamp[0]);

        // Ramp down the samples we just read.
        for (int i = 0; i < n; ++i) {
            *smps = short(float(*smps) * *ramp);
            ++smps;
            ++ramp;
        }
        reset();
	if (Restart) {
	    long mix;
	    short* t = Tempbuf;
	    short* s = Samples;

	    Playing = YES;
	    JustStopped = NO;

	    // Put the samples in <Tempbuf> this time.
	    read = readsamples_aux(num, NO);

	    // Now mix the ramped down previous buffer with the start of the
	    // sound. 
	    for (int i = 0; i < n; ++i) {
		mix = *t + *s;

		// Limit the signal to within the 16-bit range.
		if (mix < -32768)
		    mix = -32768;
		else if (mix > 32767)
		    mix = 32767;

		// Just stuff the mix in <Samples>
		*s = short(mix);
		++s;
		++t;
	    }
	}
    }
    return read;
}


//*****************************************************************************

// Read a buffer's worth of samples from the sound file. <num> is the number
// of frames to request.  Returns the number of frames read.  Called by
// soundlist:fillbuffers(). 
// If <samples> is YES, we read into <Samples> else we read into <Tempbuf>

long sound::readsamples_aux(long num, BOOL samples)
{
    long read;
    long frames_read;
    static const int siz = Read_frames * 2 * sizeof(short);
    short* storage = (samples == YES ? Samples : Tempbuf);

    // Got to the end of the sound? 
    if (Frames_left <= 0)
        return 0;
    read = (Frames_left > num) ? num : Frames_left;
    
    // If this is going to be the last read, we won't fill the buffer so set
    // all the bytes to zero first so that there's only silence at the end of
    // the buffer (instead of what was last read). 
    if (read < num)
        bzero((void *)storage, siz);

    // Get the samples. 
    frames_read = fread((void *)storage, sizeof(short) * Channels, read, Fp);
    if (frames_read < read)
        cerr << "Warning: short read of sound " << Path << "\n\r";

#ifdef LITTLE_ENDIAN
    swap_buf(storage);
#endif

    Frames_left -= frames_read;
    return frames_read;
}


//*****************************************************************************

#ifdef SGI

// Check to see if other audio processes are running and if they are issue a
// warning and don't set the global audio settings.  If nothing else is
// running, set the global output sampling rate, and each subsequent time,
// check to see that the new sound file has the same sampling rate as the
// rate we set for the first sound.  If it doesn't, signal an error (all
// sounds must have the same sampling rate if they are to be played
// correctly). Called by sound::sound(). <path> is the name of the sound and
// <srate> is the sampling rate of this sound. 

void sound::setoutputrate(long srate)
{
    long pvbuf[4];          // The buffer we pass to ALgetparams(). 
    long output_rate;       // The audio hardware sampling rate. 
    static BOOL warn = YES; // Switch to make sure that we only warn once. 

    // Check to see if other audio processes are running: 
    pvbuf[0] = AL_OUTPUT_COUNT;
    pvbuf[2] = AL_MONITOR_CTL;
    if (ALgetparams(AL_DEFAULT_DEVICE, pvbuf, 4) < 0)
        if (oserror() == AL_BAD_DEVICE_ACCESS)
            error("sound::setoutputrate", "Could not access audio hardware.");
    if ((pvbuf[1] == 0) && (pvbuf[3] == AL_MONITOR_OFF)) {
        
        // Set Sampling_rate if we haven't already. 
        if (Sampling_rate == 0) {
            output_rate = GetOutputRate();
            if (output_rate != srate) {
                output_rate = srate;
                Sampling_rate = output_rate;
                pvbuf[0] = AL_OUTPUT_RATE;
                pvbuf[1] = output_rate;

                // No other audio process is running so it's safe to
                // set global output sampling rate. 
                ALsetparams(AL_DEFAULT_DEVICE, pvbuf, 2);
            }
        }
    }
    else if (warn) {
        fprintf(stderr, "\n\r*** Warning ***  Another audio application is "
                "running.  \n\rNot setting global audio rates.\n\n\r");
        warn = NO;
    }
}


//*****************************************************************************

// Return Audio Hardware output sampling rate. Called by
// sound::setoutputrate(). 

long sound::GetOutputRate(void)
{
    long buf[4];

    buf[0] = AL_OUTPUT_RATE;
    buf[2] = AL_DIGITAL_INPUT_RATE;
    ALgetparams(AL_DEFAULT_DEVICE, buf, 4);

    // output rate is in Hz -- return it 
    if (buf[1] > 0)
        return (buf[1]);

    // Output rate is logical -- track down what it means 
    else {
        if (buf[1] == AL_RATE_AES_1) {
            
            // We are clocked off of digital input.  Find real input rate, if
            // system supports this ability. Otherwise, return
            // AL_RATE_UNDEFINED 
            if (ALgetdefault(AL_DEFAULT_DEVICE, AL_DIGITAL_INPUT_RATE) >= 0)
              return (buf[3]);
        }
        else if (buf[1] == AL_RATE_INPUTRATE)
            return (GetInputRate());
        return (AL_RATE_UNDEFINED);
    }
}


//*****************************************************************************

// Return Audio Hardware input sampling rate.  Called by GetOutputRate(). 

long sound::GetInputRate(void)
{
    long buf[6];

    buf[0] = AL_INPUT_RATE;
    buf[2] = AL_INPUT_SOURCE;
    buf[4] = AL_DIGITAL_INPUT_RATE;
    ALgetparams(AL_DEFAULT_DEVICE, buf, 6);

    // We are clocked off digital input.  Find real input rate, if possible. 
    if ((buf[1] == AL_RATE_AES_1) || (buf[3] == AL_INPUT_DIGITAL)) {
        if (ALgetdefault(AL_DEFAULT_DEVICE, AL_DIGITAL_INPUT_RATE) >= 0)
            return (buf[5]);
    }

    // Input rate is Hz and we're using an analog input 
    else if (buf[1] > 0)
        return (buf[1]);
    return (AL_RATE_UNDEFINED);
}


#endif  // SGI 

//*****************************************************************************

// EOF sound.cc 
 
