//*****************************************************************************
//
// File:          soundlist.cc
// Purpose:       soundlist is a class containing a single linked list of sound
//                objects.  The class functions manipulate this list with the
//                intention of returning a buffer of mixed samples.  The add()
//                function adds a sound to the list; the tidy() function looks
//                at every sound in the list and removes them if their
//                <Playing> field is NO; fillbuffers() fills every buffer
//                in the list with new samples, or sets the <Playing> field to
//                NO if no  more samples are to be had;
//                mix_and_limit() and simple_mix() both mix all the  
//                sound object's samples and return a pointer to the mix (see
//                below for caveats regarding these functions).  The other
//                private functions are for use by the public functions, mainly
//                fillbuffers().  
//                <Silence> is a buffer's worth of zero bytes
//                to be returned by mix() when there are no sounds in the list;
//                <Tempmixbuf> is a buffer's worth
//                of longs that we write the mix into to give us some overhead
//                and to avoid wraparound of too large sample values;
//                <Mixbuf> is the buffer we return from mix().
// Author:        Michael Edwards
// Date:          12th November 1995
// Last revision: None
//
//*****************************************************************************

#include <iostream.h>
#ifdef SGI
#include <bstring.h>
#endif
#ifdef NEXT
#include <string.h>
#endif
#include "mix.h"


//*****************************************************************************

// Constructor:  Allocate and initialize <Silence> to all zero bytes, and
// <Head> to a NULL pointer.  Also allocate storage for <Mixbuf> and
// <Tempmixbuf>.  <Silence> is never filled with samples, 

soundlist::soundlist(void)
{
    long rfx2 = sound::getRead_frames() * 2;

    Head = NULL;
    Silence = new short[rfx2];

    bzero((void *)Silence, sizeof(short) * rfx2);
    Mixbuf = new short[rfx2];
    Tempmixbuf = new long[rfx2];

    Sounds_playing = 0;
}


//*****************************************************************************

// Destructor.  Free allocated resources. 

soundlist::~soundlist(void)
{
    delete[] Silence;
    Silence = NULL;
    delete[] Mixbuf;
    Mixbuf = NULL;
    delete[] Tempmixbuf;
    Tempmixbuf = NULL;
}


//*****************************************************************************

// The mixing algorithm used here is one suggested by Doug Cooke at SGI. We
// loop through the sounds in the list and add all their samples to those
// stored in <Tempmixbuf> (which is always initialized to contain all zero
// bytes before the mixing starts, thus causing the first trip through the
// loop to be just a copying of the first sound's samples). <Tempmixbuf> is a
// buffer's worth of longs so that if the addition of two samples results in
// overflowing a 16 bit short, we don't wrap around from 32767 to -32768, or
// vice-versa.  Instead, once all the sounds have been mixed, we go through
// the samples in <Tempmixbuf> once more and check to see if any are out of
// the range of a short.  If they are, we limit them to the upper or lower
// limit accordingly.  This is still not ideal, but better than wraparound. 
// The idea is that the user can change the global amplitude in the
// key-mappings file, or with MIDI volume on the fly if any distortions are
// heard.  We don't want to hard code some scaling here as it could get very
// messy. 
// Called by audio_process(). Returns a pointer to <Mixbuf>. 

short *soundlist::mix_and_limit(void)
{
    // The next sound in the list. 
    soundlist_element *sndlist_elmt;
    // Sample we're at in <Tempmixbuf> 
    register long *current_long_sample;
    // Sample we're at in <Mixbuf>. 
    register short *current_short_sample;
    // Pointer to samples in the list. 
    register short *sampleptr;
    register long i;
    static const long rfx2 = sound::getRead_frames() * 2;

    if (Head == NULL)	              // No sounds in the list. 
	return Silence;
    else if (Head->Next == NULL)      // Only one sound in the list. 
	return (Head->Snd->getSamples());

    // Else do the mix: 
    
    // Get rid of the samples from the last call to mix()--we don't want to
    // mix our present buffer with the last buffer's samples. 
    bzero((void *)Tempmixbuf, sizeof(long) * rfx2);
    for (sndlist_elmt = Head; sndlist_elmt != NULL;
	 sndlist_elmt = sndlist_elmt->Next) {
	sampleptr = sndlist_elmt->Snd->getSamples();

	// (re)set mix to the beginning. 
	current_long_sample = Tempmixbuf;
	for (i = 0; i < rfx2; ++i) {

	    // Add the samples and store as a long: 
	    *current_long_sample += *sampleptr;
	    ++current_long_sample;
	    ++sampleptr;
	}
    }
    current_long_sample = Tempmixbuf;

    // Go to the start of <Mixbuf>. 
    current_short_sample = Mixbuf;
    
    // Limit the samples so that they fit within the range of a 16-bit short,
    // and count how many times we had to do this. 
    for (i = 0; i < rfx2; ++i) {
	if (*current_long_sample < -32768) {
	    *current_short_sample = -32768;
	    ++Limits;
	}
	else if (*current_long_sample > 32767) {
	    *current_short_sample = 32767;
	    ++Limits;
	}
	else
	  *current_short_sample = (short)*current_long_sample;
	++current_short_sample;
	++current_long_sample;
    }
    return Mixbuf;
}


//*****************************************************************************

// This is a naughty function.  Instead of mixing into a virgin buffer we're
// just going to mix into the first sound's buffer.  We can guarantee that
// the samples in the first sound's buffer won't be needed again because
// simple_mix() will be called just before sending the samples to the audio
// port, and after that, the buffer will be filled with new samples read from
// the sound file.  This saves us from copying the first sound's samples into
// the mix buffer. 
//
// The mixing algorithm is the simplest possible--just add the two samples
// together.  According to Doug Cook at SGI, for good mixing we should
// probably store the result of the addition into a buffer of a larger data
// size (e.g. int), then, when all the mixing has been done, step through the
// buffer and see if any of the samples have gone over the limit of a short. 
// If one has, it should be limited to either 32767 or -32768, whichever value
// the sample went beyond.  The way mixing is done here, if we overflow the
// upper limit, the sample will wrap around to a negative number causing
// massive discontinuity of the signal. Obviously, this isn't good, but as
// this is a real-time implementation, the speed of the algorithm is of
// paramount importance so we won't do such checks.  Instead we will rely on
// the user's ability to control gain with the MIDI fader box and the global
// amplitude scaler that can be set in the initialization file.  The point is
// to rehearse with the mixer and find acceptable levels that won't cause
// distortion of the signal, adjusting as necessary with the faders during a
// performance. 
//
// Also, assumed here is that all buffers contain stereo data--we don't want to
// do checks to see whether we should increment the sample or not. Mono
// sounds should have been made stereo elsewhere, before the mixing routine
// is called. 

short *soundlist::simple_mix(void)
{
    // The next sound in the list. 
    soundlist_element *snd;
    // Sample we're at in the first sound (mix) buffer. 
    short *current_sample;
    // Pointer to samples in the list. 
    short *sampleptr;
    short *mixptr;
    register long i;
    static const long rfx2 = sound::getRead_frames() * 2;

    if (Head == NULL)		   // No sounds in the list. 
	return Silence;
    else if (Head->Next == NULL)   // Only one sound in the list. 
	return Head->Snd->getSamples();

    // Else do the mix: 

    // Cache the address of the first sound buffer's samples: 
    mixptr = Head->Snd->getSamples();

    // Start the loop with the second sound buffer: 
    for (snd = Head->Next; snd != NULL; snd = snd->Next) {
	sampleptr = snd->Snd->getSamples();
	current_sample = mixptr;        // (re)set mix to the beginning. 
	for (i = 0; i < rfx2; i++)
	    *current_sample++ += *sampleptr++;
    }
    return mixptr;
}


//*****************************************************************************

// Add a new sound to the front of the list. Called by mix::run(). 

void soundlist::add(sound *new_sound)
{
    // Allocate a new soundlist_element. 
    soundlist_element *new_element = new soundlist_element;

    // Store the data. 
    new_element->Snd = new_sound;

    // Make the <Next> field point to the first element in the list. 
    new_element->Next = Head;

    // Now make the new element the first in the list. 
    Head = new_element;
    ++Sounds_playing;
}


//****************************************************************************

// Go through the list and remove any sounds whose <Playing> field is NO.
// Readjust the pointer to the next element so that we bypass the defunct
// one.  Called by audio_process(). 

void soundlist::tidy(void)
{
    soundlist_element *element;
    // Cache pointer to the element we want to remove from the list. 
    soundlist_element *destroy;
    
    // We have to test that element and element->Next are not NULL because
    // the update clause is executed after the body of the loop (dummy!). 
    for (element = Head; element != NULL && element->Next != NULL;
	 element = element->Next) {
	if (element->Next->Snd->getPlaying() == NO) {
	  if (element->Next->Snd->JustStopped == NO) {

	    // next element is defunct.
	    --Sounds_playing;
	    destroy = element->Next;
	    element->Next->Snd->reset();
	    element->Next = element->Next->Next;

	    // Free memory for the element we want to remove. 
	    delete destroy;
	    destroy = NULL;
	  }
	  else element->Next->Snd->JustStopped = NO;
	}
    }

    // Do the test for the first in the list.  Has to be done last. 
    if (Head != NULL && Head->Snd->getPlaying() == NO) {
      if (Head->Snd->JustStopped == NO) {
	--Sounds_playing;
	destroy = Head;
	Head->Snd->reset();
	Head = Head->Next;
	delete destroy;
	destroy = NULL;
      }
      else Head->Snd->JustStopped = NO;
    }
}


//*****************************************************************************

// Stop all the sounds that are currently playing.

void soundlist::stopall(void)
{
    soundlist_element *element;

    for (element = Head; element != NULL; element = element->Next)
	element->Snd->setPlaying(NO);
}


//*****************************************************************************

// Print the key character of all the sounds that are currently playing.

void soundlist::printkeys(void)
{
    soundlist_element *element;

    for (element = Head; element != NULL; element = element->Next) {
	putc(element->Snd->getKey(), stderr);
	putc(' ', stderr);
    }
}


//*****************************************************************************

// Loop through the sound list and fill the sound objects' sample buffers with
// new samples.  This is where we do the scaling, mono-to-stereo conversion,
// and the sampling rate conversion, wherever necessary.  Called by
// audio_process(). 

void soundlist::fillbuffers(void)
{
    soundlist_element *element;
    sound *snd;

    for (element = Head; element != NULL; element = element->Next) {
	snd = element->Snd;
	if (snd->getSrc() == 1.0) {
	    if ((snd->readsamples()) == 0) {
		snd->setPlaying(NO);
		
		// Go back to the top of the loop because there weren't any
		// samples to be had. 
		continue;
	    }
	    // Scale and reverse (if necessary) the sounds before converting
	    // mono to stereo to save a whole lot of multiplications. 
	    if (snd->getReverse()) {
		long seek = snd->getFrames_left() - sound::getRead_frames();

		snd->reverse_buffer();
		if (seek < 0)
		    seek = snd->getStart_frame() * snd->getChannels()
		      * sizeof(short);
		else
		    seek = (snd->getStart_frame() + seek)
		      * snd->getChannels() *
		      sizeof(short);
		if (fseek(snd->getFp(), seek, SEEK_SET) != 0) {
		    fprintf(stderr, "\rCouldn't seek frame for %s\n\r",
			    snd->getPath());
		    fflush(stderr);
		    
		    // If we can't position the file pointer all hell will
		    // break loose so we should stop this sound. 
		    snd->setPlaying(NO);
		}
	    }
	}
	else
	    error("soundlist::fillbuffers",
		  "Sampling-rate conversion not yet implemented.");
	snd->scale();
	if (snd->getChannels() == 1)
	    snd->mono_to_stereo();
    }
}

//*****************************************************************************

// EOF soundlist.cc 
 
